#!/usr/bin/python

from dnastorage.codec.base_codec import BaseCodec
from dnastorage.codec import base_conversion

class HuffmanTableBase:
    
    class Node:
        def __init__(self, nbase, symbol, weight=None, childlist=[]):
            self._nbase = nbase
            self.symbol = symbol
            self.weight = weight
            self.enc = None
            self._childlist = childlist
            if weight==None and len(childlist)==0:
                self.weight = -1.0
            elif weight==None:
                self.weight = sum([ x.weight for x in childlist]) 
               
        def __lt__(self, other):
            return self.weight < other.weight
        def __le__(self, other):
            return self.weight <= other.weight
        def __gt__(self, other):
            return self.weight > other.weight
        def __ge__(self, other):
            return self.weight >= other.weight
        def __eq__(self,other):
            return self.weight == other.weight
        def __ne__(self,other):
            return self.weight != other.weight
                        
        def dec(self,val,symLut):
            #print val[0],symLut,self.enc,self.symbol,symLut,self._childlist
            if self.symbol != None:
                return len(self.enc),self.symbol
            if len(val) == 0:
                return 0,None
            if len(self._childlist) <= symLut[val[0]]:
                return 0,None            
            # if len(val)==1:
            #     newval = ""
            # else:
            #     newval = val[1:]
            #print val
            return self._childlist[ symLut[val[0]] ].dec( val[1:], symLut )


        def _insert(self, enc, val, nbase, symLut):
            # figure if leaf or not
            #print enc,val
            if len(enc)==1:
                n = HuffmanTableBase.Node(nbase,val,1.0,[])
                self._childlist.append( n )
                return n
            else:
                if symLut[ enc[0] ] >= len(self._childlist):
                    self._childlist.append( HuffmanTableBase.Node(nbase,None,None,[]) )
                return self._childlist[ symLut[enc[0]] ]._insert(enc[1:],val,nbase,symLut)
                    
                    
        def symbol(self):
            return self.symbol

        def __str__(self):
            s = ""
            if len(self._childlist):
                for c in self._childlist:
                    s += "->"+str(c)

            #return "{}:({},{})".format(self.enc,self.symbol,self.weight)
            return "{}".format(self.symbol) + s

        def contains(self,sym):
            if self.symbol == sym:
                return True
            else:
                for c in self._childlist:
                    if c.contains(sym):
                        return True
                return False

        def assign_enc(self, bsyms, enc):
            self.enc = enc
            #print self._childlist
            if len(self._childlist) > 0:
                self._enc = enc
                for s,n in zip(bsyms,self._childlist):
                    n.assign_enc(bsyms,enc+s)
            #else:
            #    print str(self)

        def prevent_all_ones(self, nsyms):
            if len(self._childlist) == nsyms:
                if len(self._childlist[nsyms-1]._childlist)==0:
                    new = HuffmanTableBase.Node(self._nbase,None,None,[self._childlist[nsyms-1]])
                    self._childlist[nsyms-1] = new
                else:
                    self._childlist[nsyms-1].prevent_all_ones(nsyms)
                
            
            
    def __init__(self, nbase, base_syms, symbols, weights=None, prevent_ones=False):
        """ Build a huffman encoder/decoder table under the following assumptions:    """
        """     nbase: number of symbols in codeword (4 for DNA, 2 for binary)        """
        """     base_syms: symbols used in codeword (['A','C','G','T'] or ['0', '1']) """
        """     symbols: the symbols we'll replace with huffman codewords             """
        """     weights: frequency of the symbols (same order as symbols)             """
        self._nbase = nbase
        self._base_syms = base_syms
        self._base_syms_lookup = { b : i for i,b in enumerate(base_syms) }
        self._symbols = symbols
        self._weights = weights
        self._prevent_ones = prevent_ones
        self._nodes = []
        self._enc_table = None
        assert nbase == len(base_syms)
        self.memoize = {}
        if symbols != None:
            # we may be initialized with no symbols, in that case, don't do this
            # work to setup the huffman table, just wait
            if weights == None:
                self._weights = [ 1.0 / len(symbols) for _ in range(len(symbols)) ]
            else:
                W = sum(weights)
                #print W
                self._weights = [ float(x) / W for x in weights ]
            #print self._weights
            assert len(self._symbols) == len(self._weights)
            for s,w in zip(self._symbols,self._weights):
                n = HuffmanTableBase.Node(nbase,s,weight=w)
                self._nodes.append( n )
            self._nodes.sort(key=lambda n: n.weight)
        else:
            self.root = HuffmanTableBase.Node(self._nbase,None,None,[])


    def _insert(self, enc, val):
        n = self.root._insert(enc,val,self._nbase,self._base_syms_lookup)
        self._nodes.append(n)
        
    @classmethod
    def from_raw_table_hack(cls, table, nbase, base_syms, prevent_ones=False):
        vals = [ t[1] for t in table ]
        weights = [ 1.0 / nbase**t[0] for t in table ]
        return cls(nbase, base_syms, vals, weights, prevent_ones)

    @classmethod
    def from_raw_table(cls, table, nbase, base_syms, prevent_ones=False):
        ht = cls(nbase,base_syms,None,prevent_ones=prevent_ones)
        i = 0
        prev_l = table[0][0]
        for l,val in table:
            if l!=prev_l:
                i = (i) << (l - prev_l)
                prev_l = l
            enc = "{:0{w}b}".format(i,w=l)
            #print i,enc,val
            ht._insert(enc,val)
            i += 1
        #print ht.root
        ht.root.assign_enc(base_syms,"")
        #print ht.root
        return ht


    def get_raw_table(self,length_only=False):
        table = []
        for n in self._nodes:
            if n.symbol != None:
                if length_only:
                    table.append( [len(n.enc), n.symbol] )
                else:
                    table.append( [n.enc, n.symbol] )
                                        
        table.sort(key=lambda x: x[0])
        return table

    def get_tables(self):
        table = []
        for n in self._nodes:
            if n.symbol != None:
                table.append( [n.enc, n.symbol] )

        table.sort(key=lambda x: x[1] )

        enc = { x[1] : x[0] for x in table }
        dec = { x[0] : x[1] for x in table } 

        return (enc, dec)

    def average_length(self):
        L = 0.0
        for n in self._nodes:
            if n.symbol != None:
                L += n.weight * len(n.enc)
        return L

    def histogram(self):
        h = {}
        #H = {}
        for n in self._nodes:
            if n.symbol != None:
                l = len(n.enc)
                h[l] = h.get(l,0) + 1
                #H[l] = H.get(l,[]) + [ n.enc ]
                #H[l].sort()
        return h

    def _build_tree(self):
        _queue = []
        for n in self._nodes:
            _queue.append( n )
        _queue.sort(key=lambda n: n.weight)

        if len(_queue)==1:
            nodes = []
            nodes.append( _queue.pop(0) )
            new = HuffmanTableBase.Node(self._nbase,None,None,nodes)
            _queue.append(new)
        else:
            while len(_queue) > 1:
                nodes = []
                i = 0
                while len(_queue) > 0 and i < self._nbase:
                    nodes.append( _queue.pop(0) )
                    i += 1
                new = HuffmanTableBase.Node(self._nbase,None,None,nodes)
                self._nodes.append(new)
                _queue.append(new)
                # Note, this is ineffecient.  We should replace this dumb sort
                # with something more efficient like a min-heap
                _queue.sort(key=lambda n: n.weight)

        assert len(_queue) == 1
        r = _queue[0]

        if self._prevent_ones:
            r.prevent_all_ones(self._nbase)
        
        r.assign_enc(self._base_syms,"")
        self.root = _queue[0]

        
    def encode(self, val):
        if self._enc_table == None:
            table = []
            for n in self._nodes:
                if n.symbol != None:
                    table.append( [n.enc, n.symbol] )

            table.sort(key=lambda x: x[1])
            self._enc_table = { x[1] : x[0] for x in table }
        return self._enc_table[val]

    def decode(self, val):
        if self.memoize.has_key(val):
            return self.memoize[val]        
        ans = self.root.dec(val,self._base_syms_lookup)
        self.memoize[val] = ans
        return ans

class HuffmanTable(HuffmanTableBase):                
    def __init__(self, nbase, base_syms, symbols, weights=None,prevent_ones=False):
        HuffmanTableBase.__init__(self,nbase,base_syms,symbols,weights,prevent_ones)
        if symbols != None:
            self._build_tree()


class ErrorWhileDecodingTable(HuffmanTableBase):
    def __init__(self, nbase, base_syms, symbols=None, weights=None, prevent_ones=False):
        HuffmanTableBase.__init__(self,nbase,base_syms,symbols,weights,prevent_ones)

    def decode(self, val):
        return 0,0

    def encode(self, val):
        assert False and "Never use this for encoding"

    def average_length(self):
        return -1

    def histogram(self):
        return {}
            
class LengthLimitedHuffmanTable(HuffmanTableBase):
    # based on package-merge algorithm
    def __init__(self, L, nbase, base_syms, symbols, weights=None,prevent_ones=False):
        assert nbase==2
        assert nbase**L >= len(symbols)
        HuffmanTableBase.__init__(self,nbase,base_syms,symbols,weights=weights,prevent_ones=prevent_ones)
        self._original_nodes = self._nodes[:]
        merge = []
        new_nodes = []
        for _ in range(L,0,-1):            
            merge = new_nodes + self._nodes[:]
            #print "L = {}".format(_)
            merge.sort(key = lambda n: n.weight)
            even = len(merge)//2*2
            # drop last packet if length of merge is odd
            package = [ [merge[i],merge[i+1]] for i in range(0,even,2) ] 
            new_nodes = []
            for p in package:
                new = HuffmanTable.Node(self._nbase,None,weight=None,childlist=p)
                new_nodes.append(new)        
        merge.sort(key = lambda n: n.weight)
        code_length = []
        # adjust the weights based on code lengths
        for n in self._nodes:
            n.weight = 1/(nbase**sum([ m.contains(n.symbol) for m in merge ]))
            #print n
        # build Huffman code
        self._build_tree()
        

if __name__ == "__main__":
    import random
    R = random.Random()
    l = [ HuffmanTable.Node(2,R.randint(0,256),R.random()) for _ in range(256) ] 
    l.sort()
    
    syms = [ x for x in range(256) ]
    w = [ 1.0/256 for x in range(256) ]
    for i in range(ord('A'),ord('z'),1):
        w[i] = w[i]*2.0

    ht5 = LengthLimitedHuffmanTable(9, 2, ['0','1'], syms, w, True)
    print (ht5.average_length())
    print (ht5.histogram())


    rt = ht5.get_raw_table(True)
    print (rt)
    ht6 = HuffmanTable.from_raw_table(rt,2,['0','1'])

    print( ht6.get_raw_table())

    for x,y in zip(ht5.get_raw_table(),ht6.get_raw_table()):
        if x!=y:
            print (x,y)
    
    if ht5.get_raw_table(True) == ht6.get_raw_table(True):
        print ("They match ")

    



from dnastorage.exceptions import AllowAll

class BaseCodec(object):
    ''' Abstract Codec object to inherit to make other codecs '''
    def __init__(self,CodecObj=None,Policy=None):
        self._Obj = CodecObj
        if Policy is None:
            self._Policy = AllowAll()
        else:
            self._Policy = Policy
    def _encode(self, s):
        return s
    def encode(self,s):
        if self._Obj != None:
            return self._encode(self._Obj.encode(s))
        else:
            return self._encode(s)
    def _decode(self, s):
        return s
    def decode(self,s):
        s = self._decode(s)
        if self._Obj != None:
            return self._Obj.decode(s)
        else:
            return s



from dnastorage.codec.phys import CombineCodewords, PrependSequence, AppendSequence, InsertMidSequence
from dnastorage.codec.commafreecodec import CommaFreeCodewords
from dnastorage.codec.strand import ReedSolomonInnerCodec
from dnastorage.codec.block import ReedSolomonOuterCodec, BlockToStrand
from dnastorage.codec.layered import LayeredEncoder, LayeredDecoder

def reverse_complement(seq):
    if len(seq)==0:
        return seq
    complement = {'T':'A', 'G':'C', 'C':'G', 'A':'T'}
    r = [complement[x] for x in seq]
    r.reverse()
    return "".join(r)

def customize_RS_CFC8(is_enc,pf,primer5,primer3,intraBlockIndex=1,\
                      interBlockIndex=2,innerECC=2,strandSizeInBytes=15,\
                      blockSizeInBytes=15*185,Policy=None,\
                      withCut=None,outerECCStrands=None,minIndex=0):
    assert blockSizeInBytes % strandSizeInBytes == 0
    payload=strandSizeInBytes
    blockStrands = blockSizeInBytes / strandSizeInBytes
    if outerECCStrands is None:
        outerECCStrands = 255-blockStrands # strands
    else:
        assert outerECCStrands + blockStrands <= 255
    assert blockStrands + outerECCStrands < 256

    index = intraBlockIndex + interBlockIndex

    blockCodec = ReedSolomonOuterCodec(packetSize=blockSizeInBytes,\
                                       errorSymbols=outerECCStrands,payloadSize=payload,Policy=Policy)

    blockToStrand = BlockToStrand(payload,(blockStrands+outerECCStrands)*payload,Policy=Policy,\
                                  intraIndexSize=intraBlockIndex,\
                                  interIndexSize=interBlockIndex,filterZeroes=True)
    
    # take index into account
    # better to just say number of error symbols
    strandCodec = ReedSolomonInnerCodec(innerECC,Policy=Policy)

    codewords = CommaFreeCodewords(payload+innerECC+index,Policy=Policy)

    if not (withCut is None):
        cut = InsertMidSequence(withCut)
    else:
        cut = None
    
    pre = PrependSequence(primer5,isPrimer=True,Policy=Policy,CodecObj=cut)
    app = AppendSequence(reverse_complement(primer3), CodecObj=pre, isPrimer=True, \
                         Policy=Policy)    
    physCodec = app

    if is_enc==True:
        enc = LayeredEncoder(pf,blockSizeInBytes=blockSizeInBytes,\
                             strandSizeInBytes=strandSizeInBytes,\
                         blockCodec=blockCodec,\
                         blockToStrandCodec=blockToStrand,\
                         strandCodec=strandCodec,\
                         strandToCodewordCodec=codewords,\
                         codewordToPhysCodec=CombineCodewords(),\
                             physCodec=physCodec,Policy=Policy,minIndex=minIndex)
        return enc
    else:
        dec = LayeredDecoder(pf,blockSizeInBytes=blockSizeInBytes,\
                             strandSizeInBytes=strandSizeInBytes,\
                             blockIndexSize=interBlockIndex,\
                             blockCodec=blockCodec,\
                             strandCodec=strandCodec,\
                             physCodec=physCodec,\
                             physToStrandCodec=codewords,\
                             strandToBlockCodec=blockToStrand,Policy=Policy,minIndex=minIndex)
        return dec




bases = ['A', 'C', 'G', 'T']

values = { 'A' : 0 ,
           'C' : 1 ,
           'G' : 2 ,
           'T' : 3   }

def randomTernary(length):
    return "".join([ bases[x%3] for x in range(length) ])

def convertTernaryHelper(dec,s):
    m = int(dec % 3)
    q = int(dec / 3)
    s = s + bases[m]
    if q > 0:
        return convertTernaryHelper(q,s)
    else:
        return s

def convertTernary(dec,length):
    s = convertTernaryHelper(dec,'')
    s = s.ljust(length,bases[0])
    return s

def convertQuarnaryHelper(dec,s):
    m = int(dec % 4)
    q = int(dec / 4)
    s = s + bases[m]
    if q > 0:
        return convertQuarnaryHelper(q,s)
    else:
        return s

def convertQuarnary(dec,length):
    s = convertQuarnaryHelper(dec,'')
    s = s.ljust(length,bases[0])
    return s

def convertBaseHelper(base : int, dec : int, s : str, symbols = bases):
    m = int(dec % base)
    q = int(dec / base)
    #print(s)
    s = s + symbols[m]
    #print(s)
    if q > 0:
        return convertBaseHelper(base,q,s,symbols)
    else:
        return s
    
def convertBytetoBinary(x,length):
    s=convertBytetoBinaryHelper(x,'')
    s=s.ljust(length,"0")
    #print s
    return s

def convertBytetoBinaryHelper(x,s): #convert byte x to a binary string
    binary=['0','1']
    m=int(x%2)
    q=int(x/2)
    s=s+binary[m]
    if q >0:
        return convertBytetoBinaryHelper(q,s)
    else:
        return s

def convertBase(base, dec, length, symbols=bases):
    assert base <= len(symbols)
    s = convertBaseHelper(base,dec,'', bases)
    s = s.ljust(length,bases[0])
    return s

def convertToAnyBase(base : int, dec : int, length: int, symbols=bases):
    assert base == len(symbols)
    s = convertBaseHelper(base,dec,'', symbols)
    while len(s) < length:
        s += symbols[0]
    assert length == len(s)
    return s

def convertFromBase(base,s):
    val = 0
    power = 1
    for i in range(len(s)):
        val += values[s[i]]*power
        power *= base
    return val

def convertIntToBytes(val,num_bytes):
    val = int(val)
    if val == None:
        return [-1 for _ in range(num_bytes)]
    else:
        l = [(val & (0xff << pos*8)) >> pos*8 for pos in range(num_bytes)]
        return l

def convertBytesToInt(l):
    _sum = 0
    for i,val in enumerate(l):
        _sum += val * (256**i)
    return _sum




from random import randint
import editdistance as ed

from dnastorage.codec.base_codec import BaseCodec
from dnastorage.codec import base_conversion
from dnastorage.util.stats import stats
import dnastorage.exceptions as err

#In this order, no codewords have a hamming distance
#smaller than 2, and only two consecutive codewords have a hamming
#distance of 2. This helps make sure that consecutive indices vary by
#at least two positions

cfc_all = ['CGTGAGCA', 'ACACGATG', 'ATAGCTGC', 'ATAGAGAT', 'CAGTCATC',
           'TCATGCAT', 'TCGACGCA', 'CTATGTCG', 'ACGAGTCA', 'CACATGCG',
           'ATGTGTGC', 'TGTGACAG', 'TCATAGCA', 'TCAGATAG', 'ACGATCTC',
           'CGTATATA', 'CAGACTGA', 'CTCAGCTA', 'CTACTACA', 'ATGAGAGT',
           'ACTGTGCT', 'CATATGAC', 'GTCAGAGT', 'GTACGCAC', 'CTAGACTC',
           'TACGTGTC', 'CATAGAGT', 'TCGAGTCG', 'ATCATGCG', 'CGAGTGTG',
           'GCTCTCTA', 'TCGTATAC', 'TGTCTGCA', 'GTGTATCG', 'TCACATAC',
           'ATCAGACA', 'CACAGCGA', 'CTACTGAC', 'ATCATGTC', 'GCGTGATA',
           'ACGAGTGC', 'TCGAGATG', 'CGTAGTCA', 'ATACATCT', 'GTGACGCG',
           'CAGTCACG', 'CTCGTGCG', 'GCAGAGTG', 'CTCGCGCA', 'TCACTGTC',
           'GCTCTGAG', 'ATCAGTCG', 'ATCGAGCA', 'TATCATCT', 'CTCACTGA',
           'CTGTATAT', 'ATCGACAG', 'ATCAGCGA', 'ACTGTATG', 'GTAGTGTG',
           'CGTGTGCT', 'CGCGACAG', 'GCAGATAT', 'CTCGTACA', 'GCTCGCAT',
           'CTGTGCAT', 'AGTAGCGC', 'CTACTGCG', 'CTCGCACT', 'TCTCGCTA',
           'CTATGCAC', 'GTCAGCTG', 'CTATCTGA', 'GCAGATCG', 'GAGACTCT',
           'ACTGCACT', 'GCGATGCT', 'TGAGTAGA', 'CTAGATAT', 'GTACGCTA',
           'TCATACAC', 'GTAGTATA', 'TACGTACT', 'CGTAGACG', 'GCAGACAG',
           'TGCGAGAT', 'ACATAGAT', 'GTCTACAG', 'GTCTCTGA', 'GTGAGTCA',
           'CATGACAT', 'CGACTCTG', 'CGATCACG', 'CAGCGTAC', 'GTCTCTCT',
           'ATGTATGA', 'ACGTCGTC', 'TCGAGTAC', 'TCGATCGC', 'GTCTAGCT',
           'GCATCTCT', 'ATAGCGCA', 'CATATGCT', 'CACAGTAT', 'AGTAGCTG',
           'TACGACAT', 'ACTAGAGC', 'GACACGCT', 'CTATGTGC', 'AGACTGCG',
           'TCGTGCAC', 'GTGAGTGC', 'CACACGTG', 'GAGAGATG', 'ATCGTGCT',
           'CACAGTCG', 'ACACACTC', 'GTGATCTC', 'GTCAGTAT', 'TATCACGT',
           'ACTGTAGC', 'CTCGACTA', 'CAGACGCT', 'CAGAGATC', 'TACTGATA',
           'CATGCTGC', 'ATGTGATA', 'CGACGATG', 'GCTCTGCT', 'TGCGAGCA',
           'ATCACAGA', 'GCTCGCTG', 'GCTCTGTC', 'GCATCTGA', 'AGACTACA',
           'CGATCATC', 'TAGTAGAG', 'GTCATGAT', 'CACATGTC', 'ATGTGACT',
           'CTGTAGTC', 'GTGTATGC', 'CATGCACG', 'GCAGCTCA', 'GCAGCACG',
           'GTCGATCT', 'CTGTGTAC', 'GAGACATC', 'CGTGATGC', 'CAGCTCTA',
           'ATGAGCTG', 'TAGTGACA', 'GCATGTGC', 'GCGTAGCT', 'TACGAGCT',
           'ACGACTGA', 'GACGACGC', 'CTACTCTC', 'CACACAGA', 'CACATGAT',
           'ACGTAGCA', 'ATAGCGTC', 'ACTGCTCA', 'TATCTGCT', 'CTGTACAC',
           'CGTAGAGA', 'CAGTGATA', 'CTGTACTA', 'ACATGACT', 'TACGTGCA',
           'TGTCTGAT', 'ACACGAGC', 'GCAGATGC', 'TAGTAGCA', 'TCACATCT',
           'ACGTACAG', 'TGTATAGA', 'TCAGTACA', 'GTCGCGTC', 'CGACGTCT',
           'CACTACGC', 'TAGTGTAC', 'GACGACAG', 'TATAGCAC', 'CTATGCTA',
           'CTGTGCGC', 'ACGTATGC', 'GACACTGA', 'CGATAGAT', 'CGACACTC',
           'CTAGATGC', 'CAGCGCAT', 'TCGTGTCA', 'CACACATC', 'TCATATCG',
           'ATCAGATG', 'CACATATA', 'TCTCGATG', 'GTCTGCTA', 'AGTGAGCG',
           'CGTGAGAT', 'CGCGATCT', 'CGAGTATA', 'GTAGTGCA', 'CTACGAGA',
           'CAGAGCTA', 'TCAGCGTC', 'GTGTAGCA', 'TCGACGTG', 'CATGTGTC',
           'TCTAGAGT', 'CAGATCTG', 'CGTATCTC', 'TGTATCTA', 'TCAGTGCG',
           'ACGATATA', 'TGAGAGCG', 'GCATGCTA', 'ATCGATAT', 'GAGAGCAC',
           'CACTACAG', 'AGACTAGC', 'CAGTGTCG', 'AGTAGTCT', 'GCAGACTC',
           'TACTCTGA', 'TATGTAGA', 'CTGTAGAG', 'CGATACGC', 'GCGAGAGT',
           'TCGCGCTG', 'TCACACGC', 'TAGTGATG', 'CACTAGCG', 'CATATCTA',
           'TCATAGTC', 'CTATCTCT', 'GTGAGACG', 'TGACGCTG', 'TCTCGAGC',
           'TCAGCACT', 'ATAGTACA', 'ACACAGCA', 'CTACTAGC', 'CTAGACAG',
           'CAGTGCAC', 'CTAGTGTC', 'GCGTGACT', 'TGTGAGCT', 'GCAGCGCT',
           'CAGTGACT', 'GCAGAGCA', 'GCGCGCTA', 'TCGACATC', 'TATAGCTA',
           'GTAGTACT']

cfc_extra = ['TACGATGA', 'CACTGCTG', 'CGTAGATC', 'TCACTCTA']
cfc_special_header = 'ATCGATGC'

cfc = cfc_all[0:256]

cfc_inv = {}
def create_cfc_inv():
    global cfc_inv
    for i,c in enumerate(cfc):
        cfc_inv[c] = i
    cfc_inv[None] = -1000
    
class CommaFreeCodewords(BaseCodec):
    def __init__(self,numberSymbols,CodecObj=None,Policy=None):
        BaseCodec.__init__(self,CodecObj,Policy=Policy)
        self._numberBytes = numberSymbols
        global create_cfc_inv
        create_cfc_inv()

    def _encode(self,strand):
        enc_strand = [ cfc[s] for s in strand ]
        return enc_strand

    def _decode_cfc(self, s):
        global cfc_inv
        if s in cfc_inv:
            return cfc_inv[s]
        return None

    def _decode(self,s):
        dec = self._decode_helper(s)
        return dec

    def exact_vote(self, s):
        exact = []        
        for i in range(len(s)):
            exact.append(self._decode_cfc(s[i:i+8]))
        return exact

    def inexact_vote(self, s):
        #print s
        global cfc
        D = {}
        for x in cfc:
            d = ed.eval(s,x)
            D[d] = D.get(d,[]) + [x]
        if 0 in D:
            assert len(D[0])==1
            return [100,D[0][0],0]        
        elif 1 in D:
            if len(D[1])==1:
                return [100,D[1][0],8]
            elif len(D[1])>1:
                p = D[1][ randint(0,len(D[1])-1) ]
                return [1.0/len(D[1])*100,p,8]
        elif 2 in D:
            if len(D[2])==1:
                return [100,D[2][0],8]
            elif len(D[2])>1:
                p = D[2][ randint(0,len(D[2])-1) ]
                return [1.0/len(D[2])*100,p,8]
        else:
                return [0.0,None,8]


    def get_next_exact(self, i, exact):
        while i < len(exact):
            if exact[i] != None:
                return i
            i += 1
        return i

    def _decode_helper(self, s):
        numSyms = self._numberBytes
        exact = self.exact_vote(s)

        #print len(s)
        
        new_strand = []
        i = 0
        while i < len(s) and len(new_strand) < numSyms:
            #print i
            if not(exact[i] is None):
                if i%8 != 0:
                    stats.inc("CFC8::misAlignedCodeword")
                new_strand.append(exact[i])
                #print i,"->",exact[i]
                i += 8
            else:
                e = err.DNABadCodeword("Missing expected CFC8 symbol")
                if self._Policy.allow(e):
                    j = self.get_next_exact(i,exact)
                    skipped = int(round((j-i)/8.0))
                    i_tmp = i
                    for k in range(skipped):
                        #new_strand.append(-1)
                        v = self.inexact_vote(s[i_tmp:i_tmp+8])
                        if v[0]==100:
                            new_strand.append(cfc_inv[v[1]])
                            i_tmp+=v[2]
                        else:
                            i_tmp+=8
                            new_strand.append(-1)
                    i = j
                else:
                    raise e
        
        if len(new_strand) != numSyms:
            e = err.DNAStrandPayloadWrongSize("Payload wrong size in CFC8")
            stats.inc("CFC8:payloadWrongSize")
            if self._Policy.allow(e):            
                if len(new_strand) < numSyms:
                    while len(new_strand) < numSyms:
                        new_strand.append(-1)
                else:
                    new_strand = new_strand[0:numSyms]
            else:
                raise e


        cnt=0
        found=False
        for _ in new_strand:
            stats.inc("CFC8::TotalCodewords")
            if _ == -1:
                found=True
                stats.inc("CFC8::UnidentifiedCodeword")
            else:
                stats.inc("CFC8::IdentifiedCodeword")
        if found:
            #print "found error"
            stats.inc("CFC8::strandsWithErrors")
            stats.inc("CFC8::strandsTotal")
        else:
            stats.inc("CFC8::strandsTotal")
            
        return new_strand
                
            # if j<3:
            #     i += (j-i)
            #     continue
            # v = inexact_vote(s[i:i+10])
            # if v[0] == 100:
            #     new_strand.append(v[1])
            #     if j < len(s):
            #         i += (j-i)
            #     else:
            #         i += v[0]
            # else:
            #     i += 1
    



from math import log, ceil
from collections import Counter

from dnastorage.exceptions import *
from dnastorage.codec.base_codec import *
from dnastorage.codec import base_conversion
from dnastorage.codec.reedsolomon.rs import ReedSolomon,get_reed_solomon,ReedSolomonError
from dnastorage.util.stats import stats

import logging
logger = logging.getLogger('dna.storage.codec.block')
logger.addHandler(logging.NullHandler())

block_count = 0

def reportBlockStatus(blocks,bindex,interIndexSize,intraIndexSize):
    global block_count
    for j,b in enumerate(blocks):
        block_count+=1
        contents = [ '_' for i in range(256) ]
        for s in b[1]:
            intra = base_conversion.convertBytesToInt(s[interIndexSize:interIndexSize+intraIndexSize])
            if intra == -1:
                continue
            logger.debug("intra = {}".format(intra))
            assert intra >=0 and intra <= 255
            contents[intra] = '*'        
        stats["reportBlockStatus({})::block_size({}<{}>)".format(bindex,b[0],block_count)] = len(b[1])
        stats["reportBlockStatus({})::block_profile({}<{}>)".format(bindex,b[0],block_count)] = "".join(contents)


def partitionStrandsIntoBlocks(strands, interIndexSize=2):
    logger.debug("partitionStrandsIntoBlocks:strands={}".format(len(strands)))
    blocksD = {}
    for s in strands:
        idx = base_conversion.convertBytesToInt(s[0:interIndexSize])
        blocksD[idx] = blocksD.get(idx,[]) + [s]
    blocks = blocksD.items()
    return blocks

def doMajorityVote(strands, indexBytes=3):
    key_value = {}
    strand_array = []
    for s in strands:
        key = base_conversion.convertBytesToInt(s[0:indexBytes])
        if key is not None:
            key_value[key] = key_value.get(key,[]) + [s]
            
    for key in key_value:
        data=[]
        if len(key_value[key])==1:
            strand_array.append(key_value[key][0])
            continue

        mx = max([len(data_strand) for data_strand in key_value[key]])
        for x in range(0,mx):
            #get a list of values that belong to the same location
            # if there's a smaller strand, due to error, this assert
            # will fire before attempting to access an illegal index
            same_position_values = []
            for data_strand in key_value[key]:
                if x < len(data_strand):
                    same_position_values.append(data_strand[x])

            cnt=Counter(same_position_values)
            most_common_value=cnt.most_common(1)[0][0]
            data.append(most_common_value)
            #put the key and data into an array
        strand_array.append(data)
        
    return strand_array


class NormalizeBlock(BaseCodec):
    def __init__(self, blockSizeInBytes):
        self.blockSize = blockSizeInBytes
        def _encode(self, packet):
            if len(packet[0]) < self.blockSize:
                block = packet[0] + [0] * (len(packet[0])-self.blockSize)
            else:
                block = packet[0]
            return packet[0],block

        def _decode(self, packet):
            return packet

class BlockToStrand(BaseCodec):
    """ This class encodes an (index,block) pair into a strand that contains 
        it's own index follwed by the data.  After this conversion, the index
        simpy appears as part of the strand.
        The _encode returns a list of strands.
    
        The _decode function expects to receive a set of strands that belong to
        the same block. Then they are assembled into a cohesive block and returned.     
    """

    def __init__(self, strandSizeInBytes, blockSizeInBytes, intraIndexSize=1, interIndexSize=2, nSyms=256, CodecObj=None, Policy=None, filterZeroes=False):
        super(BlockToStrand,self).__init__(CodecObj=CodecObj,Policy=Policy)
        self._strandSizeInBytes = strandSizeInBytes
        self._blockSize = blockSizeInBytes
        self._interIndex = interIndexSize
        self._intraIndex = intraIndexSize
        self._nSyms = nSyms
        self._removeZeroes = filterZeroes
        logger.info("filterZeroes = {}".format(filterZeroes))
        
    def allzeroes(self, l):
        for z in l:
            if z != 0:
                return False
        return True

    def _filter_zeroes(self, block):
        if self._removeZeroes == False:
            return [ False for _ in range(0,len(block),self._strandSizeInBytes) ]

        zeroes = []
        for i in range(0,len(block),self._strandSizeInBytes): 
            strand = block[i:i+self._strandSizeInBytes]
            if self.allzeroes(strand):
                zeroes.append(True)
            else:
                zeroes.append(False)
                    
        fil = [ False for _ in range(len(zeroes)) ]
        assert len(fil) == len(zeroes)
        for i,z in enumerate(zeroes):
            if i > 0 and i < len(zeroes)-1:
                if zeroes[i-1]==True and zeroes[i+1]==True and zeroes[i]==True:
                    fil[i] = True
        return fil
                    
    def _encode(self, packet):
        strands = []
        bindex = base_conversion.convertIntToBytes(packet[0],self._interIndex)
        block = packet[1]

        fil = self._filter_zeroes(block)            
        assert len(bindex) <= self._interIndex
        
        for i in range(0,len(block),self._strandSizeInBytes):
            #if allzeroes(block[i:i+self._strandSizeInBytes]):
            #    continue
            if fil[int(i/self._strandSizeInBytes)]==True: # don't emit zeroes
                stats.inc("BlockToStrand::filterAllZeroStrand")
                continue
            
            sindex = base_conversion.convertIntToBytes(i/self._strandSizeInBytes,self._intraIndex)
            assert len(sindex) <= self._intraIndex
            s = bindex + sindex  + block[i:i+self._strandSizeInBytes]            
            strands.append(s)
            
        return strands

    def _decode(self, packet):        
        """ accepts a list of strands and coverts them into a contiguous block """

        d = {} # track all strands we find

        stats.inc("BlockToStrand::_decode::numberOfBlocks")
        stats.append("BlockToStrand::_decode::indices",packet[0])
        
        prior_bindex = packet[0]
        strands = packet[1]
        max_index = 0
        for s in strands:
            bindex = base_conversion.convertBytesToInt(s[0:self._interIndex])
            sindex = base_conversion.convertBytesToInt(s[self._interIndex:self._interIndex+self._intraIndex])

            if bindex != prior_bindex:
                e = DNABlockBadIndex("Bad index {}!={}".format(bindex,prior_bindex))
                #print "Bad index {}!={}".format(bindex,prior_bindex)
                if self._Policy.allow(e):
                    # ignore this strand
                    if sindex in d:
                        # already found a strand with this sindex, don't use this one
                        continue
                    else:
                        pass
                else:
                    raise e

            if len(s[self._interIndex+self._intraIndex:]) != self._strandSizeInBytes:            
                # handle exception condition here!
                err = DNABlockPayloadWrongSize("Strand payload size should be {} bytes but was {}.".format(self._strandSizeInBytes,len(s[self._interIndex+self._intraIndex:])))
                if self._Policy.allow(err):
                    if len(s[self._interIndex+self._intraIndex:]) > self._strandSizeInBytes:
                        d[sindex] = s[self._interIndex+self._intraIndex:self._interIndex+self._intraIndex+self._strandSizeInBytes]
                    else:
                        d[sindex] = s[self._interIndex+self._intraIndex:]+[-1]*(self._strandSizeInBytes-len(s[self._interIndex+self._intraIndex:]))
                else:
                    raise err
            else:
                d[sindex] = s[self._interIndex+self._intraIndex:]
                
        data = []
        max_index = int(self._blockSize / self._strandSizeInBytes)
        contents = [ '_' for _ in range(max_index) ]
        for i in range(max_index):
            if not i in d:
                if self._removeZeroes:
                    if not (i-1 in d) and not ((i+1) in d):
                        # guess that it's all zeros
                        stats.inc("BlockToStrand::decode::guessAllZeroStrand")
                        data += [ 0 for _ in range(self._strandSizeInBytes) ]
                        contents[i] = '0'
                    elif (i-1) in d and self.allzeroes(d[i-1]) and not (i+1 in d):
                        # guess that it's all zeros
                        stats.inc("BlockToStrand::decode::guessAllZeroStrand")
                        data += [ 0 for _ in range(self._strandSizeInBytes) ]
                        contents[i] = '0'
                    elif (i+1) in d and self.allzeroes(d[i+1]) and not (i-1 in d):
                        # guess that it's all zeros
                        stats.inc("BlockToStrand::decode::guessAllZeroStrand")
                        data += [ 0 for _ in range(self._strandSizeInBytes) ]
                        contents[i] = '0'
                    else:
                        err = DNABlockMissingIndex("Block missing index = {}".format(i))
                        stats.inc("BlockToStrand::decode::guessMissingStrand")
                        if self._Policy.allow(err):
                            #print "Block missing index = {}".format(i)
                            data += d.get(i,[-1 for _ in range(self._strandSizeInBytes)])
                            contents[i] = 'M'
                        else:
                            raise err
                else:
                    err = DNABlockMissingIndex("Block missing index = {}".format(i))
                    if self._Policy.allow(err):
                        #print "Block missing index = {}".format(i)
                        data += d.get(i,[-1 for _ in range(self._strandSizeInBytes)])
                    else:
                        raise err
            else:
                contents[i] = '-'
                data += d[i]
        stats.unique("BlockToStrand::_decode::block_profile({})".format(prior_bindex),"".join(contents))
                
        # smaller is allowed due to end of file, larger is a potential problem
        if len(data) > self._blockSize:
            err = DNABlockTooLargeError("Block too large, is {} should be {}".format(len(data),self._blockSize))
            if self._Policy.allow(err):                
                data = data[:self._blockSize]
            else:
                raise err

        #print data
        #stats.append("BlockToStrand::_decode::data[{}]".format(prior_bindex),data)
        
        return prior_bindex,data
        

class DoNothingOuterCodec(BaseCodec):
    """This is a simple do nothing Outer Codec to bypass this step when testing new designs"""
    def __init__(self,packetSize,payloadSize,CodecObj=None,Policy=None):
        BaseCodec.__init__(self,CodecObj=CodecObj,Policy=Policy)
        self._packetSize=packetSize
        self._payloadSize=payloadSize

    def _encode(self,packet):
        """simply check length of packet to make sure it is a multiple of payload size"""
        index = packet[0]
        data = packet[1][:]       
        assert len(packet[1])<=self._packetSize

        # hopefully the last packet if it's not a full packetSize
        # so, pad it out with zeros to make an even number of strands if
        # isn't already        
        if len(data) < self._packetSize:
            stats.inc("DoNothingOuterCodec.padPacket")
            # normalize to multiple of payloadSize
            rem = len(data) % self._payloadSize
            data += [0]*(self._payloadSize-rem)

        #print "Data length {}".format(len(data))
        assert len(data)==self._packetSize
        return packet[0],data

        
class ReedSolomonOuterCodec(BaseCodec):
    """ReedSolomonOuterCodec takes a block of data as input and produces a new
    block of data that's error encoded. 

    The encoding input is a linear array of length payloadSize * nCol bytes,
    where nRow is ideally set to the payload per strand. However, it's
    not a strict requirement.

    Bcol must be less than rs.field_charac, which is either 2**8 or 2**16.

    """
    def __init__(self,packetSize,errorSymbols,payloadSize,c_exp=8,CodecObj=None,Policy=None):
        BaseCodec.__init__(self,CodecObj=CodecObj,Policy=Policy)

        self._rs = get_reed_solomon(c_exp=c_exp)

        self._packetSize = packetSize
        self._errorSymbols = int(errorSymbols)
        self._payloadSize = payloadSize
        
        self._lengthMessage = packetSize / payloadSize + errorSymbols

        assert(self._lengthMessage <= self._rs.field_charac) # requirement of RS
        
    """
    packet is a tuple with the key at position 0 and value at position 1. This should return
    the Reed-Solomon encoded byte array.
    """
    def _encode(self,packet):

        index = packet[0]
        data = packet[1][:]
        
        assert packet[0] < 256**3
        assert len(packet[1])<=self._packetSize

        # hopefully the last packet if it's not a full packetSize
        # so, pad it out with zeros to make an even number of strands if
        # isn't already
        
        if len(data) < self._packetSize:
            stats.inc("RSOuterCodec.padPacket")
            # normalize to multiple of payloadSize
            rem = len(data) % self._payloadSize
            data += [0]*(self._payloadSize-rem)

        #print data
        assert len(data) % self._payloadSize == 0
        rows = len(data) / self._payloadSize
        # construct message

        ecc = []        
        for i in range(self._payloadSize):
            message = data[i:len(data):self._payloadSize]
            mesecc = self._rs.rs_encode_msg(message, self._errorSymbols)
            ecc += mesecc[len(message):]
            #print "encode message=",mesecc
            #print message, ecc
            #print  len(message), len(ecc), self._errorSymbols
            #assert len(ecc) == self._errorSymbols

        tranpose = []
        for i in range( int(self._errorSymbols) ):
            syms = ecc[i:len(ecc):self._errorSymbols]
            data += syms
                    
        # separate out key and value
        # convert mesecc into a string
        return packet[0],data

    """
    This function expects a list of unsigned integers in GF(256). For now, erasures are
    denoted with -1.
    """
    def _decode(self,packet):
        #print packet[1]
        data = [x for x in packet[1]]
        #print data
        rows = len(data) / self._payloadSize
        for i in range(self._payloadSize):
            message = data[i:len(data):self._payloadSize]            
            erasures = [ j for j in range(len(message)) if message[j]==-1 ]
            try:
                # correct the message
                corrected_message, corrected_ecc = \
                    self._rs.rs_correct_msg(message, \
                                            self._errorSymbols, \
                                            erase_pos=erasures)
                #debug help
                #print corrected_message
                #print corrected_ecc
                data[i:len(data):self._payloadSize] = corrected_message + corrected_ecc
                stats.inc("RSOuterCodec::correct")
            except ReedSolomonError as e:
                #print "couldn't correct block {}".format(message)
                stats.inc("RSOuterCodec::ReedSolomonError")
                # wrap exception into a library specific one when checking policy:
                wr_e = DNAReedSolomonOuterCodeError(msg=\
                             "RSOuterCodec found error at index={}".format(packet[0]))
                if self._Policy.allow(wr_e):
                    # fix -1
                    corrected_message = [ max(0,_) for _ in message ]
                    data[i:len(data):self._payloadSize] = corrected_message
                    pass # nothing to do
                else:
                    # policy doesn't allow handling, raise exception
                    raise wr_e
                # just proceed without further error correction
                pass
            except Exception as e:
                if self._Policy.allow(e):
                    pass
                else:
                    raise e

        # discard outer correction codes
        data = data[0:len(data)-self._errorSymbols*self._payloadSize]        
        return packet[0],data
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###################################################
### Universal Reed-Solomon Codec
### initially released at Wikiversity
### Creative Commons License: Attribution-ShareAlike 3.0
###
### 2019/7/19: jmt modified to be first class data type  
###
###################################################

################### INIT and stuff ###################

try: # compatibility with Python 3+
    xrange
except NameError:
    xrange = range

class ReedSolomonError(Exception):
    pass

################### GALOIS FIELD ELEMENTS MATHS ###################

def rwh_primes1(n):
    # http://stackoverflow.com/questions/2068372/fastest-way-to-list-all-primes-below-n-in-python/3035188#3035188
    ''' Returns  a list of primes < n '''
    sieve = [True] * (n/2)
    for i in xrange(3,int(n**0.5)+1,2):
        if sieve[i/2]:
            sieve[i*i/2::i] = [False] * ((n-i*i-1)/(2*i)+1)
    return [2] + [2*i+1 for i in xrange(1,n/2) if sieve[i]]

class ReedSolomon:
    def __init__(self, generator=2, c_exp=8, prim=0x11d, errs=10 ):
        # For efficiency, gf_exp[] has size 2*GF_SIZE, so that a
        # simple multiplication of two numbers can be resolved without
        # calling % 255. For more infos on how to generate this
        # extended exponentiation table, see paper: "Fast software
        # implementation of finite field operations", Cheng Huang and
        # Lihao Xu, Washington University in St. Louis, Tech. Rep
        # (2003).
        if prim is None:
            prim = self.find_prime_polys(generator=generator, c_exp=c_exp,  \
                                         fast_primes=True, single=True)
            print ("prim=0x{:x}".format(prim))
            
        self.init_tables(prim=prim,generator=generator,c_exp=c_exp)
        #self.gen = self.rs_generator_poly(errs, 0, generator)
        
    def find_prime_polys(self, generator=2, c_exp=8, fast_primes=False, single=False):
        '''Compute the list of prime polynomials for the given generator and
           galois field characteristic exponent.
        '''

        # fast_primes will output less results but will be significantly
        # faster.  single will output the first prime polynomial found, so
        # if all you want is to just find one prime polynomial to generate
        # the LUT for Reed-Solomon to work, then just use that.

        # A prime polynomial (necessarily irreducible) is necessary to
        # reduce the multiplications in the Galois Field, so as to avoid
        # overflows.

        # Why do we need a "prime polynomial"? Can't we just reduce modulo
        # 255 (for GF(2^8) for example)? Because we need the values to be
        # unique.

        # For example: if the generator (alpha) = 2 and c_exp = 8 (GF(2^8)
        # == GF(256)), then the generated Galois Field (0, 1, a, a^1, a^2,
        # ..., a^(p-1)) will be galois field it becomes 0, 1, 2, 4, 8, 16,
        # etc. However, upon reaching 128, the next value will be doubled
        # (ie, next power of 2), which will give 256. Then we must reduce,
        # because we have overflowed above the maximum value of 255. But,
        # if we modulo 255, this will generate 256 == 1. Then 2, 4, 8, 16,
        # etc. giving us a repeating pattern of numbers. This is very bad,
        # as it's then not anymore a bijection (ie, a non-zero value
        # doesn't have a unique index). That's why we can't just modulo
        # 255, but we need another number above 255, which is called the
        # prime polynomial.

        # Why so much hassle? Because we are using precomputed look-up
        # tables for multiplication: instead of multiplying a*b, we
        # precompute alpha^a, alpha^b and alpha^(a+b), so that we can just
        # use our lookup table at alpha^(a+b) and get our result. But just
        # like in our original field we had 0,1,2,...,p-1 distinct unique
        # values, in our "LUT" field using alpha we must have unique
        # distinct values (we don't care that they are different from the
        # original field as long as they are unique and distinct). That's
        # why we need to avoid duplicated values, and to avoid duplicated
        # values we need to use a prime irreducible polynomial.

        # Here is implemented a bruteforce approach to find all these
        # prime polynomials, by generating every possible prime
        # polynomials (ie, every integers between field_charac+1 and
        # field_charac*2), and then we build the whole Galois Field, and
        # we reject the candidate prime polynomial if it duplicates even
        # one value or if it generates a value above field_charac (ie,
        # cause an overflow).

        # Note that this algorithm is slow if the field is too big (above
        # 12), because it's an exhaustive search algorithm. There are
        # probabilistic approaches, and almost surely prime approaches,
        # but there is no determistic polynomial time algorithm to find
        # irreducible monic polynomials. More info can be found at:
        # http://people.mpi-inf.mpg.de/~csaha/lectures/lec9.pdf

        # Another faster algorithm may be found at Adleman, Leonard M.,
        # and Hendrik W. Lenstra. "Finding irreducible polynomials over
        # finite fields." Proceedings of the eighteenth annual ACM
        # symposium on Theory of computing. ACM, 1986.

        # Prepare the finite field characteristic (2^p - 1), this also
        # represent the maximum possible value in this field
        self.root_charac = 2 # we're in GF(2)
        self.field_charac = int(self.root_charac**c_exp - 1)
        field_charac_next = int(self.root_charac**(c_exp+1) - 1)

        prim_candidates = []
        if fast_primes:
            # generate maybe prime polynomials and check later if they
            # really are irreducible
            prim_candidates = rwh_primes1(field_charac_next)
            # filter out too small primes
            prim_candidates = [x for x in prim_candidates if x > self.field_charac] 
        else:
            # try each possible prime polynomial, but skip even numbers
            # (because divisible by 2 so necessarily not irreducible)
            prim_candidates = xrange(field_charac+2, field_charac_next, root_charac)

        # Start of the main loop
        correct_primes = []
        # try potential candidates primitive irreducible polys
        for prim in prim_candidates:
            # memory variable to indicate if a value was already generated
            # in the field (value at index x is set to 1) or not (set to 0
            # by default)
            seen = [0] * (self.field_charac+1) 
            conflict = False # flag to know if there was at least one conflict

            # Second loop, build the whole Galois Field
            x = 1
            for i in xrange(self.field_charac):
                # Compute the next value in the field (ie, the next power
                # of alpha/generator)
                x = self.gf_mult_noLUT(x, generator, prim, self.field_charac+1)

                # Rejection criterion: if the value overflowed (above
                # field_charac) or is a duplicate of a previously
                # generated power of alpha, then we reject this polynomial
                # (not prime)
                if x > self.field_charac or seen[x] == 1:
                    conflict = True
                    break
                # Else we flag this value as seen (to maybe detect future
                # duplicates), and we continue onto the next power of alpha
                else:
                    seen[x] = 1

            # End of the second loop: if there's no conflict (no overflow
            # nor duplicated value), this is a prime polynomial!
            if not conflict: 
                correct_primes.append(prim)
                if single: return prim

        # Return the list of all prime polynomials. You can use the
        # following to print the hexadecimal representation of each prime
        # polynomial: print [hex(i) for i in correct_primes]
        return correct_primes 

    def init_tables(self, prim=0x11d, generator=2, c_exp=8):
        '''Precompute the logarithm and anti-log tables for faster computation
        later, using the provided primitive polynomial.  These tables are
        used for multiplication/division since addition/substraction are
        simple XOR operations inside GF of characteristic 2.  The basic
        idea is quite simple: since b**(log_b(x), log_b(y)) == x * y given
        any number b (the base or generator of the logarithm), then we can
        use any number b to precompute logarithm and anti-log
        (exponentiation) tables to use for multiplying two numbers x and
        y.  That's why when we use a different base/generator number, the
        log and anti-log tables are drastically different, but the
        resulting computations are the same given any such tables.  For
        more infos, see
        https://en.wikipedia.org/wiki/Finite_field_arithmetic#Implementation_tricks
        '''
        # generator is the generator number (the "increment" that will be
        # used to walk through the field by multiplication, this must be a
        # prime number). This is basically the base of the
        # logarithm/anti-log tables. Also often noted "alpha" in academic
        # books.

        # prim is the primitive/prime (binary) polynomial and must be
        # irreducible (ie, it can't represented as the product of two
        # smaller polynomials). It's a polynomial in the binary sense:
        # each bit is a coefficient, but in fact it's an integer between
        # field_charac+1 and field_charac*2, and not a list of gf
        # values. The prime polynomial will be used to reduce the
        # overflows back into the range of the Galois Field without
        # duplicating values (all values should be unique). See the
        # function find_prime_polys() and: http://research.swtch.com/field
        # and http://www.pclviewer.com/rs2/galois.html

        # note that the choice of generator or prime polynomial doesn't
        # matter very much: any two finite fields of size p^n have
        # identical structure, even if they give the individual elements
        # different names (ie, the coefficients of the codeword will be
        # different, but the final result will be the same: you can always
        # correct as many errors/erasures with any choice for those
        # parameters). That's why it makes sense to refer to all the
        # finite fields, and all decoders based on Reed-Solomon, of size
        # p^n as one concept: GF(p^n). It can however impact sensibly the
        # speed (because some parameters will generate sparser tables).
        # c_exp is the exponent for the field's characteristic GF(2^c_exp)

        #Members of ReedSolomon class: gf_exp, gf_log, field_charac
        self.field_charac = int(2**c_exp - 1)
        self.gf_exp = [0] * (self.field_charac * 2) # anti-log (exponential) table. The first two elements will always be [GF256int(1), generator]
        self.gf_log = [0] * (self.field_charac+1) # log table, log[0] is impossible and thus unused

        # For each possible value in the galois field 2^8, we will
        # pre-compute the logarithm and anti-logarithm (exponential) of this value

        # To do that, we generate the Galois Field F(2^p) by building a
        # list starting with the element 0 followed by the (p-1)
        # successive powers of the generator a : 1, a, a^1, a^2, ...,
        # a^(p-1).
        x = 1

        for i in xrange(self.field_charac): # we could skip index 255 which is equal to index 0 because of modulo: g^255==g^0 but either way, this does not change the later outputs (ie, the ecc symbols will be the same either way)
            self.gf_exp[i] = x # compute anti-log for this value and store it in a table
            self.gf_log[x] = i # compute log at the same time
            x = self.gf_mult_noLUT(x, generator, prim, self.field_charac+1)

            # If you use only generator==2 or a power of 2, you can use
            # the following which is faster than gf_mult_noLUT():

            #x <<= 1 # multiply by 2 (change 1 by another number y to
            #multiply by a power of 2^y)

            #if x & 0x100: # similar to x >= 256, but a lot faster (because 0x100 == 256)

                #x ^= prim

                # substract the primary polynomial to the current value
                # (instead of 255, so that we get a unique set made of
                # coprime numbers), this is the core of the tables generation

        # Optimization: double the size of the anti-log table so that we
        # don't need to mod 255 to stay inside the bounds (because we will
        # mainly use this table for the multiplication of two GF numbers,
        # no more).
        for i in xrange(self.field_charac, self.field_charac * 2):
            self.gf_exp[i] = self.gf_exp[i - self.field_charac]

        return [self.gf_log, self.gf_exp]

    def gf_add(self, x, y):
        return x ^ y

    def gf_sub(self, x, y):
        return x ^ y # in binary galois field, substraction is just the same as addition (since we mod 2)

    def gf_neg(self,x):
        return x

    def gf_mul(self, x, y):
        if x == 0 or y == 0:
            return 0
        return self.gf_exp[(self.gf_log[x] + self.gf_log[y]) % self.field_charac]

    def gf_div(self, x, y):
        if y == 0:
            raise ZeroDivisionError()
        if x == 0:
            return 0
        return self.gf_exp[(self.gf_log[x] + self.field_charac - self.gf_log[y]) % self.field_charac]

    def gf_pow(self, x, power):
        return self.gf_exp[(self.gf_log[x] * power) % self.field_charac]

    def gf_inverse(self, x):
        return self.gf_exp[self.field_charac - self.gf_log[x]] # gf_inverse(x) == gf_div(1, x)

    def gf_mult_noLUT(self, x, y, prim=0, field_charac_full=256, carryless=True):
        '''Galois Field integer multiplication using Russian Peasant
        Multiplication algorithm (faster than the standard
        multiplication + modular reduction).  If prim is 0 and
        carryless=False, then the function produces the result for a
        standard integers multiplication (no carry-less arithmetics
        nor modular reduction).
        '''
        r = 0
        while y: # while y is above 0 y is odd, then add the
            # corresponding x to r (the sum of all x's corresponding
            # to odd y's will give the final product). Note that since
            # we're in GF(2), the addition is in fact an XOR (very
            # important because in GF(2) the multiplication and
            # additions are carry-less, thus it changes the result!).
            if y & 1: r = r ^ x if carryless else r + x 
            y = y >> 1 # equivalent to y // 2
            x = x << 1 # equivalent to x*2
            # GF modulo: if x >= 256 then apply modular reduction
            # using the primitive polynomial (we just substract, but
            # since the primitive number can be above 256 then we
            # directly XOR).
            if prim > 0 and x & field_charac_full: x = x ^ prim 
        return r


################### GALOIS FIELD POLYNOMIALS MATHS ###################

    def gf_poly_scale(self, p, x):
        return [self.gf_mul(p[i], x) for i in xrange(len(p))]

    def gf_poly_add(self, p,q):
        r = [0] * max(len(p),len(q))
        for i in xrange(len(p)):
            r[i+len(r)-len(p)] = p[i]
        for i in xrange(len(q)):
            r[i+len(r)-len(q)] ^= q[i]
        return r

    def gf_poly_mul(self, p, q):
        '''Multiply two polynomials, inside Galois Field (but the procedure is generic). Optimized function by precomputation of log.'''
        # Pre-allocate the result array
        r = [0] * (len(p) + len(q) - 1)
        # Precompute the logarithm of p
        lp = [self.gf_log[p[i]] for i in xrange(len(p))]
        # Compute the polynomial multiplication (just like the outer product of two vectors, we multiply each coefficients of p with all coefficients of q)
        for j in xrange(len(q)):
            qj = q[j] # optimization: load the coefficient once
            if qj != 0: # log(0) is undefined, we need to check that
                lq = self.gf_log[qj] # Optimization: precache the logarithm of the current coefficient of q
                for i in xrange(len(p)):
                    if p[i] != 0: # log(0) is undefined, need to check that...
                        r[i + j] ^= self.gf_exp[lp[i] + lq] # equivalent to: r[i + j] = gf_add(r[i+j], gf_mul(p[i], q[j]))
        return r

    def gf_poly_mul_simple(self, p, q): # simple equivalent way of multiplying two polynomials without precomputation, but thus it's slower
        '''Multiply two polynomials, inside Galois Field'''
        # Pre-allocate the result array
        r = [0] * (len(p) + len(q) - 1)
        # Compute the polynomial multiplication (just like the outer product of two vectors, we multiply each coefficients of p with all coefficients of q)
        for j in xrange(len(q)):
            for i in xrange(len(p)):
                r[i + j] ^= self.gf_mul(p[i], q[j]) # equivalent to: r[i + j] = gf_add(r[i+j], gf_mul(p[i], q[j])) -- you can see it's your usual polynomial multiplication
        return r

    def gf_poly_neg(self,poly):
        '''Returns the polynomial with all coefficients negated. In GF(2^p), negation does not change the coefficient, so we return the polynomial as-is.'''
        return poly

    def gf_poly_div(self, dividend, divisor):
        '''Fast polynomial division by using Extended Synthetic Division and
        optimized for GF(2^p) computations (doesn't work with standard
        polynomials outside of this galois field, see the Wikipedia
        article for generic algorithm).
        '''

        # CAUTION: this function expects polynomials to follow the
        # opposite convention at decoding: the terms must go from the
        # biggest to lowest degree (while most other functions here expect
        # a list from lowest to biggest degree). eg: 1 + 2x + 5x^2 = [5,
        # 2, 1], NOT [1, 2, 5]

        # Copy the dividend list and pad with 0 where the ecc bytes will
        # be computed
        msg_out = list(dividend) 
        #normalizer = divisor[0] # precomputing for performance
        for i in xrange(len(dividend) - (len(divisor)-1)):
            #msg_out[i] /= normalizer # for general polynomial division
            #(when polynomials are non-monic), the usual way of using

            # synthetic division is to divide the divisor g(x) with its
            # leading coefficient, but not needed here.
            coef = msg_out[i] # precaching log(0) is undefined, so we need
            # to avoid that case explicitly (and it's also a good
            # optimization).
            if coef != 0:
                # in synthetic division, we always skip the first
                # coefficient of the divisior, because it's only used to
                # normalize the dividend coefficient
                for j in xrange(1, len(divisor)): 
                    if divisor[j] != 0: # log(0) is undefined equivalent
                        # to the more mathematically correct (but xoring
                        # directly is faster): msg_out[i + j] +=
                        # -divisor[j] * coef
                        msg_out[i + j] ^= self.gf_mul(divisor[j], coef) 

        # The resulting msg_out contains both the quotient and the
        # remainder, the remainder being the size of the divisor (the
        # remainder has necessarily the same degree as the divisor -- not
        # length but degree == length-1 -- since it's what we couldn't
        # divide from the dividend), so we compute the index where this
        # separation is, and return the quotient and remainder.
        separator = -(len(divisor)-1)
        return msg_out[:separator], msg_out[separator:] # return quotient, remainder.

    def gf_poly_eval(self, poly, x):
        '''Evaluates a polynomial in GF(2^p) given the value for x. This is based on Horner's scheme for maximum efficiency.'''
        y = poly[0]
        for i in xrange(1, len(poly)):
            y = self.gf_mul(y, x) ^ poly[i]
        return y


################### REED-SOLOMON ENCODING ###################

    def rs_generator_poly(self, nsym, fcr=0, generator=2):
        '''Generate an irreducible generator polynomial (necessary to encode a message into Reed-Solomon)'''
        g = [1]
        for i in xrange(int(nsym)):
            g = self.gf_poly_mul(g, [1, self.gf_pow(generator, i+fcr)])
        return g

    def rs_generator_poly_all(self, max_nsym, fcr=0, generator=2):
        '''Generate all irreducible generator polynomials up to max_nsym (usually you can use n, the length of the message+ecc). Very useful to reduce processing time if you want to encode using variable schemes and nsym rates.'''
        g_all = {}
        g_all[0] = g_all[1] = [1]
        for nsym in xrange(max_nsym):
            g_all[nsym] = self.rs_generator_poly(nsym, fcr, generator)
        return g_all

    def rs_simple_encode_msg(self, msg_in, nsym, fcr=0, generator=2, gen=None):
        '''Simple Reed-Solomon encoding (mainly an example for you to understand how it works, because it's slower than the inlined function below)'''

        if (len(msg_in) + nsym) > self.field_charac: raise ValueError("Message is too long (%i when max is %i)" % (len(msg_in)+nsym, self.field_charac))
        if gen is None: gen = self.rs_generator_poly(nsym, fcr, generator)

        # Pad the message, then divide it by the irreducible generator polynomial
        _, remainder = self.gf_poly_div(msg_in + [0] * (len(gen)-1), gen)
        # The remainder is our RS code! Just append it to our original message to get our full codeword (this represents a polynomial of max 256 terms)
        msg_out = msg_in + remainder
        # Return the codeword
        return msg_out

    def rs_encode_msg(self, msg_in, nsym, fcr=0, generator=2, gen=None):
        '''Reed-Solomon main encoding function, using polynomial division
    (Extended Synthetic Division, the fastest algorithm available to my
    knowledge), better explained at http://research.swtch.com/field

        '''

        if (len(msg_in) + nsym) > self.field_charac: raise ValueError("Message is too long (%i when max is %i)" % (len(msg_in)+nsym, self.field_charac))
        if gen is None: gen = self.rs_generator_poly(nsym, fcr, generator)
        #gen = self.gen
        #print "gen={}".format(gen)
        # Init msg_out with the values inside msg_in and pad with
        # len(gen)-1 bytes (which is the number of ecc symbols).
        msg_out = [0] * (len(msg_in) + len(gen)-1)
        # Initializing the Synthetic Division with the dividend (= input
        # message polynomial)
        msg_out[:len(msg_in)] = msg_in

        # Synthetic division main loop
        for i in xrange(len(msg_in)):
            # Note that it's msg_out here, not msg_in. Thus, we reuse the
            # updated value at each iteration (this is how Synthetic
            # Division works: instead of storing in a temporary register
            # the intermediate values, we directly commit them to the
            # output).
            coef = msg_out[i]

            # log(0) is undefined, so we need to manually check for this case.
            if coef != 0:
                # in synthetic division, we always skip the first
                # coefficient of the divisior, because it's only used to
                # normalize the dividend coefficient (which is here
                # useless since the divisor, the generator polynomial, is
                # always monic)
                for j in xrange(1, len(gen)):
                    #if gen[j] != 0: # log(0) is undefined so we need to
                    #check that, but it slow things down in fact and it's
                    #useless in our case (reed-solomon encoding) since we
                    #know that all coefficients in the generator are not 0
                    msg_out[i+j] ^= self.gf_mul(gen[j], coef) # equivalent to msg_out[i+j] += gf_mul(gen[j], coef)

        # At this point, the Extended Synthetic Divison is done, msg_out
        # contains the quotient in msg_out[:len(msg_in)] and the remainder
        # in msg_out[len(msg_in):]. Here for RS encoding, we don't need
        # the quotient but only the remainder (which represents the RS
        # code), so we can just overwrite the quotient with the input
        # message, so that we get our complete codeword composed of the
        # message + code.
        msg_out[:len(msg_in)] = msg_in

        return msg_out


    ################### REED-SOLOMON DECODING ###################

    def rs_calc_syndromes(self, msg, nsym, fcr=0, generator=2):
        '''Given the received codeword msg and the number of error correcting symbols (nsym), computes the syndromes polynomial.
        Mathematically, it's essentially equivalent to a Fourrier Transform (Chien search being the inverse).
        '''
        # Note the "[0] +" : we add a 0 coefficient for the lowest degree
        # (the constant). This effectively shifts the syndrome, and will
        # shift every computations depending on the syndromes (such as the
        # errors locator polynomial, errors evaluator polynomial, etc. but
        # not the errors positions).

        # This is not necessary, you can adapt subsequent computations to
        # start from 0 instead of skipping the first iteration (ie, the
        # often seen range(1, n-k+1)),
        synd = [0] * nsym
        for i in xrange(nsym):
            synd[i] = self.gf_poly_eval(msg, self.gf_pow(generator, i+fcr))
        return [0] + synd # pad with one 0 for mathematical precision (else we can end up with weird calculations sometimes)

    def rs_correct_errata(self, msg_in, synd, err_pos, fcr=0, generator=2): # err_pos is a list of the positions of the errors/erasures/errata
        '''Forney algorithm, computes the values (error magnitude) to correct the input message.'''

        # calculate errata locator polynomial to correct both errors and
        # erasures (by combining the errors positions given by the error
        # locator polynomial found by BM with the erasures positions given
        # by caller)
        coef_pos = [len(msg_in) - 1 - p for p in err_pos]
        # need to convert the positions to coefficients degrees for the
        # errata locator algo to work (eg: instead of [0, 1, 2] it will
        # become [len(msg)-1, len(msg)-2, len(msg) -3])
        err_loc = self.rs_find_errata_locator(coef_pos, generator)
        # calculate errata evaluator polynomial (often called Omega or
        # Gamma in academic papers)
        err_eval = self.rs_find_error_evaluator(synd[::-1], err_loc, len(err_loc)-1)[::-1]

        # Second part of Chien search to get the error location polynomial
        # X from the error positions in err_pos (the roots of the error
        # locator polynomial, ie, where it evaluates to 0)
        X = [] # will store the position of the errors
        for i in xrange(len(coef_pos)):
            l = self.field_charac - coef_pos[i]
            X.append( self.gf_pow(generator, -l) )

        # Forney algorithm: compute the magnitudes
        E = [0] * (len(msg_in)) # will store the values that need to be corrected (substracted) to the message containing errors. This is sometimes called the error magnitude polynomial.
        Xlength = len(X)
        for i, Xi in enumerate(X):

            Xi_inv = self.gf_inverse(Xi)

            # Compute the formal derivative of the error locator polynomial (see Blahut, Algebraic codes for data transmission, pp 196-197).
            # the formal derivative of the errata locator is used as the denominator of the Forney Algorithm, which simply says that the ith error value is given by error_evaluator(gf_inverse(Xi)) / error_locator_derivative(gf_inverse(Xi)). See Blahut, Algebraic codes for data transmission, pp 196-197.
            err_loc_prime_tmp = []
            for j in xrange(Xlength):
                if j != i:
                    err_loc_prime_tmp.append( self.gf_sub(1, self.gf_mul(Xi_inv, X[j])) )
            # compute the product, which is the denominator of the Forney algorithm (errata locator derivative)
            err_loc_prime = 1
            for coef in err_loc_prime_tmp:
                err_loc_prime = self.gf_mul(err_loc_prime, coef)
            # equivalent to: err_loc_prime = functools.reduce(gf_mul, err_loc_prime_tmp, 1)

            # Compute y (evaluation of the errata evaluator polynomial)
            # This is a more faithful translation of the theoretical equation contrary to the old forney method. Here it is an exact reproduction:
            # Yl = omega(Xl.inverse()) / prod(1 - Xj*Xl.inverse()) for j in len(X)
            y = self.gf_poly_eval(err_eval[::-1], Xi_inv) # numerator of the Forney algorithm (errata evaluator evaluated)
            y = self.gf_mul(self.gf_pow(Xi, 1-fcr), y) # adjust to fcr parameter

            # Compute the magnitude
            magnitude = self.gf_div(y, err_loc_prime) # magnitude value of the error, calculated by the Forney algorithm (an equation in fact): dividing the errata evaluator with the errata locator derivative gives us the errata magnitude (ie, value to repair) the ith symbol
            E[err_pos[i]] = magnitude # store the magnitude for this error into the magnitude polynomial

        # Apply the correction of values to get our message corrected! (note that the ecc bytes also gets corrected!)
        # (this isn't the Forney algorithm, we just apply the result of decoding here)
        msg_in = self.gf_poly_add(msg_in, E) # equivalent to Ci = Ri - Ei where Ci is the correct message, Ri the received (senseword) message, and Ei the errata magnitudes (minus is replaced by XOR since it's equivalent in GF(2^p)). So in fact here we substract from the received message the errors magnitude, which logically corrects the value to what it should be.
        return msg_in

    def rs_find_error_locator(self, synd, nsym, erase_loc=None, erase_count=0):
        '''Find error/errata locator and evaluator polynomials with Berlekamp-Massey algorithm'''
        # The idea is that BM will iteratively estimate the error locator
        # polynomial.  To do this, it will compute a Discrepancy term
        # called Delta, which will tell us if the error locator polynomial
        # needs an update or not (hence why it's called discrepancy: it
        # tells us when we are getting off board from the correct value).

        # Init the polynomials
        if erase_loc:
            # if the erasure locator polynomial is supplied, we init with
            # its value, so that we include erasures in the final locator
            # polynomial
            err_loc = list(erase_loc)
            old_loc = list(erase_loc)
        else:
            err_loc = [1]
            # This is the main variable we want to fill, also called Sigma
            # in other notations or more formally the errors/errata
            # locator polynomial.
            old_loc = [1]
            # BM is an iterative algorithm, and we need the errata locator
            # polynomial of the previous iteration in order to update
            # other necessary variables.  L = 0 # update flag variable,
            # not needed here because we use an alternative equivalent way
            # of checking if update is needed (but using the flag could
            # potentially be faster depending on if using length(list) is
            # taking linear time in your language, here in Python it's
            # constant so it's as fast.

        # Fix the syndrome shifting: when computing the syndrome, some
        # implementations may prepend a 0 coefficient for the lowest
        # degree term (the constant). This is a case of syndrome shifting,
        # thus the syndrome will be bigger than the number of ecc symbols
        # (I don't know what purpose serves this shifting). If that's the
        # case, then we need to account for the syndrome shifting when we
        # use the syndrome such as inside BM, by skipping those prepended
        # coefficients.  Another way to detect the shifting is to detect
        # the 0 coefficients: by definition, a syndrome does not contain
        # any 0 coefficient (except if there are no errors/erasures, in
        # this case they are all 0). This however doesn't work with the
        # modified Forney syndrome, which set to 0 the coefficients
        # corresponding to erasures, leaving only the coefficients
        # corresponding to errors.
        synd_shift = 0
        if len(synd) > nsym: synd_shift = len(synd) - nsym

        for i in xrange(nsym-erase_count):
            # generally: nsym-erase_count == len(synd), except when you input
            # a partial erase_loc and using the full syndrome instead of the
            # Forney syndrome, in which case nsym-erase_count is more correct
            # (len(synd) will fail badly with IndexError).
            if erase_loc:
                # if an erasures locator polynomial was provided to init
                # the errors locator polynomial, then we must skip the
                # FIRST erase_count iterations (not the last iterations,
                # this is very important!)
                K = erase_count+i+synd_shift
            else:
                # if erasures locator is not provided, then either there's
                # no erasures to account or we use the Forney syndromes,
                # so we don't need to use erase_count nor erase_loc (the
                # erasures have been trimmed out of the Forney syndromes).
                K = i+synd_shift

            # Compute the discrepancy Delta Here is the close-to-the-books
            # operation to compute the discrepancy Delta: it's a simple
            # polynomial multiplication of error locator with the
            # syndromes, and then we get the Kth element.

            #delta = gf_poly_mul(err_loc[::-1], synd)[K] # theoretically it should be gf_poly_add(synd[::-1], [1])[::-1] instead of just synd, but it seems it's not absolutely necessary to correctly decode.

            # But this can be optimized: since we only need the Kth
            # element, we don't need to compute the polynomial
            # multiplication for any other element but the Kth. Thus to
            # optimize, we compute the polymul only at the item we need,
            # skipping the rest (avoiding a nested loop, thus we are
            # linear time instead of quadratic).  This optimization is
            # actually described in several figures of the book "Algebraic
            # codes for data transmission", Blahut, Richard E., 2003,
            # Cambridge university press.
            delta = synd[K]
            for j in xrange(1, len(err_loc)):
                delta ^= self.gf_mul(err_loc[-(j+1)], synd[K - j])
                # delta is also called discrepancy. Here we do a partial
                # polynomial multiplication (ie, we compute the polynomial
                # multiplication only for the term of degree K). Should be
                # equivalent to brownanrs.polynomial.mul_at().
            #print "delta", K, delta, list(gf_poly_mul(err_loc[::-1], synd)) # debugline

            # Shift polynomials to compute the next degree
            old_loc = old_loc + [0]

            # Iteratively estimate the errata locator and evaluator polynomials
            if delta != 0: # Update only if there's a discrepancy
                if len(old_loc) > len(err_loc): # Rule B (rule A is implicitly defined because rule A just says that we skip any modification for this iteration)
                #if 2*L <= K+erase_count: # equivalent to len(old_loc) > len(err_loc), as long as L is correctly computed
                    # Computing errata locator polynomial Sigma
                    new_loc = self.gf_poly_scale(old_loc, delta)
                    old_loc = self.gf_poly_scale(err_loc, self.gf_inverse(delta)) # effectively we are doing err_loc * 1/delta = err_loc // delta
                    err_loc = new_loc
                    # Update the update flag
                    #L = K - L # the update flag L is tricky: in Blahut's schema, it's mandatory to use `L = K - L - erase_count` (and indeed in a previous draft of this function, if you forgot to do `- erase_count` it would lead to correcting only 2*(errors+erasures) <= (n-k) instead of 2*errors+erasures <= (n-k)), but in this latest draft, this will lead to a wrong decoding in some cases where it should correctly decode! Thus you should try with and without `- erase_count` to update L on your own implementation and see which one works OK without producing wrong decoding failures.

                # Update with the discrepancy
                err_loc = self.gf_poly_add(err_loc, self.gf_poly_scale(old_loc, delta))

        # Check if the result is correct, that there's not too many errors to correct
        while len(err_loc) and err_loc[0] == 0: del err_loc[0] # drop leading 0s, else errs will not be of the correct size
        errs = len(err_loc) - 1
        if (errs-erase_count) * 2 + erase_count > nsym:
            raise ReedSolomonError("Too many errors to correct")    # too many errors to correct

        return err_loc

    def rs_find_errata_locator(self, e_pos, generator=2):
        '''Compute the erasures/errors/errata locator polynomial from the
           erasures/errors/errata positions (the positions must be
           relative to the x coefficient, eg: "hello worldxxxxxxxxx" is
           tampered to "h_ll_ worldxxxxxxxxx" with xxxxxxxxx being the ecc
           of length n-k=9, here the string positions are [1, 4], but the
           coefficients are reversed since the ecc characters are placed
           as the first coefficients of the polynomial, thus the
           coefficients of the erased characters are n-1 - [1, 4] = [18,
           15] = erasures_loc to be specified as an argument.

        '''
        e_loc = [1] # just to init because we will multiply, so it must be 1 so that the multiplication starts correctly without nulling any term
        # erasures_loc = product(1 - x*alpha**i) for i in erasures_pos and where alpha is the alpha chosen to evaluate polynomials.
        for i in e_pos:
            e_loc = self.gf_poly_mul( e_loc, self.gf_poly_add([1], [self.gf_pow(generator, i), 0]) )
        return e_loc

    def rs_find_error_evaluator(self, synd, err_loc, nsym):
        '''Compute the error (or erasures if you supply sigma=erasures locator
           polynomial, or errata) evaluator polynomial Omega from the
           syndrome and the error/erasures/errata locator Sigma.
        '''

        # Omega(x) = [ Synd(x) * Error_loc(x) ] mod x^(n-k+1)
        _, remainder = self.gf_poly_div( self.gf_poly_mul(synd, err_loc), ([1] + [0]*(nsym+1)) ) # first multiply syndromes * errata_locator, then do a
                                                                                       # polynomial division to truncate the polynomial to the
                                                                                       # required length

        # Faster way that is equivalent
        #remainder = gf_poly_mul(synd, err_loc) # first multiply the syndromes with the errata locator polynomial
        #remainder = remainder[len(remainder)-(nsym+1):] # then slice the list to truncate it (which represents the polynomial), which
                                                         # is equivalent to dividing by a polynomial of the length we want

        return remainder

    def rs_find_errors(self, err_loc, nmess, generator=2): # nmess is len(msg_in)
        '''Find the roots (ie, where evaluation = zero) of error polynomial by brute-force trial, this is a sort of Chien's search
        (but less efficient, Chien's search is a way to evaluate the polynomial such that each evaluation only takes constant time).'''
        errs = len(err_loc) - 1
        err_pos = []
        for i in xrange(nmess): # normally we should try all 2^8 possible values, but here we optimize to just check the interesting symbols
            if self.gf_poly_eval(err_loc, self.gf_pow(generator, i)) == 0: # It's a 0? Bingo, it's a root of the error locator polynomial,
                                                            # in other terms this is the location of an error
                err_pos.append(nmess - 1 - i)
        # Sanity check: the number of errors/errata positions found should be exactly the same as the length of the errata locator polynomial
        if len(err_pos) != errs:
            # couldn't find error locations
            raise ReedSolomonError("Too many (or few) errors found by Chien Search for the errata locator polynomial!")
        return err_pos

    def rs_forney_syndromes(self, synd, pos, nmess, generator=2):
        # Compute Forney syndromes, which computes a modified syndromes to compute only errors (erasures are trimmed out). Do not confuse this with Forney algorithm, which allows to correct the message based on the location of errors.
        erase_pos_reversed = [nmess-1-p for p in pos] # prepare the coefficient degree positions (instead of the erasures positions)

        # Optimized method, all operations are inlined
        fsynd = list(synd[1:])      # make a copy and trim the first coefficient which is always 0 by definition
        for i in xrange(len(pos)):
            x = self.gf_pow(generator, erase_pos_reversed[i])
            for j in xrange(len(fsynd) - 1):
                fsynd[j] = self.gf_mul(fsynd[j], x) ^ fsynd[j + 1]
            #fsynd.pop() # useless? it doesn't change the results of computations to leave it there

        # Theoretical way of computing the modified Forney syndromes: fsynd = (erase_loc * synd) % x^(n-k)
        # See Shao, H. M., Truong, T. K., Deutsch, L. J., & Reed, I. S. (1986, April). A single chip VLSI Reed-Solomon decoder. In Acoustics, Speech, and Signal Processing, IEEE International Conference on ICASSP'86. (Vol. 11, pp. 2151-2154). IEEE.ISO 690
        #erase_loc = rs_find_errata_locator(erase_pos_reversed, generator=generator) # computing the erasures locator polynomial
        #fsynd = gf_poly_mul(erase_loc[::-1], synd[1:]) # then multiply with the syndrome to get the untrimmed forney syndrome
        #fsynd = fsynd[len(pos):] # then trim the first erase_pos coefficients which are useless. Seems to be not necessary, but this reduces the computation time later in BM (thus it's an optimization).

        return fsynd

    def rs_correct_msg(self, msg_in, nsym, fcr=0, generator=2, erase_pos=None, only_erasures=False):
        '''Reed-Solomon main decoding function'''
        if len(msg_in) > self.field_charac:
            # Note that it is in fact possible to encode/decode messages that are longer than field_charac, but because this will be above the field, this will generate more error positions during Chien Search than it should, because this will generate duplicate values, which should normally be prevented thank's to the prime polynomial reduction (eg, because it can't discriminate between error at position 1 or 256, both being exactly equal under galois field 2^8). So it's really not advised to do it, but it's possible (but then you're not guaranted to be able to correct any error/erasure on symbols with a position above the length of field_charac -- if you really need a bigger message without chunking, then you should better enlarge c_exp so that you get a bigger field).
            raise ValueError("Message is too long (%i when max is %i)" % (len(msg_in), field_charac))

        msg_out = list(msg_in)     # copy of message
        # erasures: set them to null bytes for easier decoding (but this is not necessary, they will be corrected anyway, but debugging will be easier with null bytes because the error locator polynomial values will only depend on the errors locations, not their values)
        if erase_pos is None:
            erase_pos = []
        else:
            for e_pos in erase_pos:
                msg_out[e_pos] = 0
        # check if there are too many erasures to correct (beyond the Singleton bound)
        if len(erase_pos) > nsym: raise ReedSolomonError("Too many erasures to correct")
        # prepare the syndrome polynomial using only errors (ie: errors = characters that were either replaced by null byte
        # or changed to another character, but we don't know their positions)
        synd = self.rs_calc_syndromes(msg_out, nsym, fcr, generator)
        # check if there's any error/erasure in the input codeword. If not (all syndromes coefficients are 0), then just return the message as-is.
        if max(synd) == 0:
            return msg_out[:-nsym], msg_out[-nsym:]  # no errors

        # Find errors locations
        if only_erasures:
            err_pos = []
        else:
            # compute the Forney syndromes, which hide the erasures from the original syndrome (so that BM will just have to deal with errors, not erasures)
            fsynd = self.rs_forney_syndromes(synd, erase_pos, len(msg_out), generator)
            # compute the error locator polynomial using Berlekamp-Massey
            err_loc = self.rs_find_error_locator(fsynd, nsym, erase_count=len(erase_pos))
            # locate the message errors using Chien search (or bruteforce search)
            err_pos = self.rs_find_errors(err_loc[::-1], len(msg_out), generator)
            if err_pos is None:
                raise ReedSolomonError("Could not locate error")


        # Find errors values and apply them to correct the message
        # compute errata evaluator and errata magnitude polynomials, then correct errors and erasures
        msg_out = self.rs_correct_errata(msg_out, synd, (erase_pos + err_pos), fcr, generator) # note that we here use the original syndrome, not the forney syndrome
                                                             # (because we will correct both errors and erasures, so we need the full syndrome)
        # check if the final message is fully repaired
        synd = self.rs_calc_syndromes(msg_out, nsym, fcr, generator)
        if max(synd) > 0:
            raise ReedSolomonError("Could not correct message")     # message could not be repaired
        # return the successfully decoded message
        return msg_out[:-nsym], msg_out[-nsym:] # also return the corrected ecc block so that the user can check()

    def rs_correct_msg_nofsynd(self, msg_in, nsym, fcr=0, generator=2, erase_pos=None, only_erasures=False):
        '''Reed-Solomon main decoding function, without using the modified
        Forney syndromes This demonstrates how the decoding process is
        done without using the Forney syndromes (this is the most common
        way nowadays, avoiding Forney syndromes require to use a modified
        Berlekamp-Massey that will take care of the erasures by itself,
        it's a simple matter of modifying some initialization variables
        and the loop ranges)

        '''
        if len(msg_in) > self.field_charac:
            raise ValueError("Message is too long (%i when max is %i)" % (len(msg_in), field_charac))

        msg_out = list(msg_in)     # copy of message
        # erasures: set them to null bytes for easier decoding (but this is not necessary, they will be corrected anyway, but debugging will be easier with null bytes because the error locator polynomial values will only depend on the errors locations, not their values)
        if erase_pos is None:
            erase_pos = []
        else:
            for e_pos in erase_pos:
                msg_out[e_pos] = 0
        # check if there are too many erasures
        if len(erase_pos) > nsym: raise ReedSolomonError("Too many erasures to correct")
        # prepare the syndrome polynomial using only errors (ie: errors =
        # characters that were either replaced by null byte or changed to
        # another character, but we don't know their positions)

        synd = self.rs_calc_syndromes(msg_out, nsym, fcr, generator)
        # check if there's any error/erasure in the input codeword. If not
        # (all syndromes coefficients are 0), then just return the
        # codeword as-is.
        if max(synd) == 0:
            return msg_out[:-nsym], msg_out[-nsym:]  # no errors

        # prepare erasures locator and evaluator polynomials
        erase_loc = None
        #erase_eval = None
        erase_count = 0
        if erase_pos:
            erase_count = len(erase_pos)
            erase_pos_reversed = [len(msg_out)-1-eras for eras in erase_pos]
            erase_loc = self.rs_find_errata_locator(erase_pos_reversed, generator=generator)
            #erase_eval = rs_find_error_evaluator(synd[::-1], erase_loc, len(erase_loc)-1)

        # prepare errors/errata locator polynomial
        if only_erasures:
            err_loc = erase_loc[::-1]
            #err_eval = erase_eval[::-1]
        else:
            err_loc = self.rs_find_error_locator(synd, nsym, erase_loc=erase_loc, erase_count=erase_count)
            err_loc = err_loc[::-1]
            #err_eval = rs_find_error_evaluator(synd[::-1], err_loc[::-1], len(err_loc)-1)[::-1] # find error/errata evaluator polynomial (not really necessary since we already compute it at the same time as the error locator poly in BM)

        # locate the message errors
        err_pos = self.rs_find_errors(err_loc, len(msg_out), generator) # find the roots of the errata locator polynomial (ie: the positions of the errors/errata)
        if err_pos is None:
            raise ReedSolomonError("Could not locate error")

        # compute errata evaluator and errata magnitude polynomials, then correct errors and erasures
        msg_out = self.rs_correct_errata(msg_out, synd, err_pos, fcr=fcr, generator=generator)
        # check if the final message is fully repaired
        synd = self.rs_calc_syndromes(msg_out, nsym, fcr, generator)
        if max(synd) > 0:
            raise ReedSolomonError("Could not correct message")
        # return the successfully decoded message
        return msg_out[:-nsym], msg_out[-nsym:] # also return the corrected ecc block so that the user can check()

    def rs_check(self, msg, nsym, fcr=0, generator=2):
        '''Returns true if the message + ecc has no error of false otherwise (may not always catch a wrong decoding or a wrong message, particularly if there are too many errors -- above the Singleton bound --, but it usually does)'''
        return ( max(self.rs_calc_syndromes(msg, nsym, fcr, generator)) == 0 )

    
def get_reed_solomon(c_exp=8):
    if c_exp==8:
        return ReedSolomon(generator=2, c_exp=8, prim=0x18d )
    elif c_exp==16:
        return ReedSolomon(generator=2, c_exp=16, prim=0x1002d )
    else:
        assert False and "ReedSolomon codec doesn't support field of width {} yet. Use c_exp=8 or c_exp=16".format(2**c_exp)
        

if __name__ == "__main__":
    from random import randint
    rs16 = ReedSolomon(generator=2, c_exp=16, prim=0x1002d )

    rs8 = ReedSolomon(generator=2, c_exp=8, prim=0x18d )

    message = [ randint(0,255) for _ in range(15) ]

    new_msg = rs16.rs_encode_msg(message, 20)
    new_msg2 = rs8.rs_encode_msg(message, 20)
    print (message)
    print (new_msg)
    print (new_msg2)



from io import BytesIO
import sys
import logging

from dnastorage.util.packetizedfile import ReadPacketizedFilestream, WritePacketizedFilestream
import dnastorage.system.formats as formats
import dnastorage.system.header as header
from dnastorage.codec.base_conversion import convertIntToBytes, convertBytesToInt

logger = logging.getLogger('dna.storage.system.dnafile')
logger.addHandler(logging.NullHandler())

def reverse_complement(seq):
    complement = {'T':'A', 'G':'C', 'C':'G', 'A':'T'}
    r = [complement[x] for x in seq]
    r.reverse()
    return "".join(r)

class DNAFile:
    def __init__(self):
        return

    @classmethod
    def open(self, filename, op, primer5, primer3, format_name="", write_incomplete_file=False, fsmd_abbrev='FSMD',flanking_primer5="",flanking_primer3="",use_flanking_primer_for_decoding=False,use_single_primer=False,preview_mode=False,reverse_primer3_from_seq=False):
        '''        
        1. filename is the input file of strands
        2. op is the file operation: 'r' or 'w'
        3. format_name is optional for reading and required for writing. 
        4. primer5 is the coding primer, and primer3 is the non-coding primer.
        '''     
        # check if we are reading or writing
        if op=="r":
            
            with open(filename,"r") as fd:
                logger.debug("open {} for reading.".format(filename))
                strands = get_strands(fd)

                if use_flanking_primer_for_decoding==True:
                    h = header.decode_file_header(strands,flanking_primer5+primer5,\
                                       flanking_primer3+primer3,fsmd_abbrev=fsmd_abbrev)
                else:
                    h = header.decode_file_header(strands,primer5,primer3,fsmd_abbrev=fsmd_abbrev)
            
                logger.debug("decoded header: {}".format(h)) 
                assert h['version'][0] <= header.system_version['major']
                assert h['version'][1] <= header.system_version['minor']

                if h['formatid'] == formats.file_system_formatid_by_abbrev("Segmented"):
                    logger.debug("SegmentedReadDNAFile({},{},{})".format(filename,primer5,primer3))
                    return SegmentedReadDNAFile(input=filename,\
                                                primer5=primer5,primer3=primer3,\
                                                write_incomplete_file=write_incomplete_file,\
                                                fsmd_abbrev=fsmd_abbrev,\
                                                flanking_primer5=flanking_primer5,\
                                                flanking_primer3=flanking_primer3,\
                                                use_flanking_primer_for_decoding=use_flanking_primer_for_decoding,\
                                                use_single_primer=use_single_primer,\
                                                header=h,\
                                                preview_mode=preview_mode,\
                                                reverse_primer3_from_seq=reverse_primer3_from_seq)
                else:
                    return ReadDNAFile(input=filename, primer5=primer5, primer3=primer3,\
                                       fsmd_abbrev=fsmd_abbrev,\
                                       flanking_primer5=flanking_primer5,\
                                       flanking_primer3=flanking_primer3,\
                                       use_flanking_primer_for_decoding=use_flanking_primer_for_decoding,\
                                       header=h,preview_mode=preview_mode)
                
                # return ReadDNAFile(input=filename, primer5=primer5, primer3=primer3,\
                #                    fsmd_abbrev=fsmd_abbrev,\
                #                    flanking_primer5=flanking_primer5,\
                #                    flanking_primer3=flanking_primer3,\
                #                    use_flanking_primer_for_decoding=use_flanking_primer_for_decoding)
        elif "w" in op and "s" in op:
            return SegmentedWriteDNAFile(output=filename,primer5=primer5,primer3=primer3, \
                                         format_name=format_name,fsmd_abbrev=fsmd_abbrev,\
                                         flanking_primer5=flanking_primer5,\
                                         flanking_primer3=flanking_primer3)       
        elif op=="w":
            return WriteDNAFile(output=filename,primer5=primer5,primer3=primer3, \
                                format_name=format_name,fsmd_abbrev=fsmd_abbrev,\
                                flanking_primer5=flanking_primer5,\
                                flanking_primer3=flanking_primer3)
                
        # elif op=="w":
        #     return WriteDNAFile(output=filename,primer5=primer5,primer3=primer3, \
        #                         format_name=format_name,fsmd_abbrev=fsmd_abbrev,\
        #                         flanking_primer5=flanking_primer5,\
        #                         flanking_primer3=flanking_primer3)
        else:
            return None

    def flush(self):
        return

    def close(self):
        return

    def readable(self):
        assert False
    def writable(self):
        assert False

def get_strands(in_fd):
    strands = []
    while True:
        s = in_fd.readline()        
        if len(s) == 0:
            break
        s = s.strip()
        if s.startswith('%'):
            continue
        strands.append(s)
    return strands

class ReadDNAFile(DNAFile):
    # ReadDNAFile reads a set of strands from a file.  It finds the header,
    # determines compatibility and encoding type, and then decodes the file.
    #
    def __init__(self,**kwargs):     
        DNAFile.__init__(self)

        if 'input' in kwargs:
            self.input_filename = kwargs['input']
            self.in_fd = open(self.input_filename,"r")
            self.need_to_close = True
        elif 'in_fd' in kwargs:
            self.in_fd = kwargs['in_fd']
            self.input_filename = ""
            self.need_to_close = False
            
        if not ('fsmd_abbrev' in kwargs):
            self.fsmd_abbrev = 'FSMD'
        else:
            self.fsmd_abbrev = kwargs['fsmd_abbrev']
            
        assert 'primer5' in kwargs and 'primer3' in kwargs
        self.primer5 = kwargs['primer5']
        self.primer3 = kwargs['primer3']

        if 'flanking_primer5' in kwargs:
            self.flanking_primer5 = kwargs['flanking_primer5']
        else:
            self.flanking_primer5 = ''

        if 'flanking_primer3' in kwargs:
            self.flanking_primer3 = kwargs['flanking_primer3']
        else:
            self.flanking_primer3 = ''

        if 'use_flanking_primer_for_decoding' in kwargs:
            self.use_flanking_primers = kwargs['use_flanking_primer_for_decoding']
        else:
            self.use_flanking_primers = False
            
        # get all the strands ( this will be bad for large files )
        strands = get_strands(self.in_fd)
        if 'header' in kwargs and kwargs['header'] != None:
            h = kwargs['header']
        else:
            h = header.decode_file_header(strands,self.primer5,self.primer3,fsmd_abbrev=self.fsmd_abbrev)    

        self.strands = header.pick_nonheader_strands(strands,self.primer5)
        
        assert h['version'][0] <= header.system_version['major']
        assert h['version'][1] <= header.system_version['minor']

        self.header = h 
        self.formatid = h['formatid']
        self.size = h['size']

        # set up mem_buffer 
        self.mem_buffer = BytesIO()
        
        if self.formatid == 0x1000:
            # let sub-classes handle initialization
            return

        dec_func = formats.file_system_decoder(self.formatid)
            
            
        self.mem_buffer = BytesIO()
        self.pf = WritePacketizedFilestream(self.mem_buffer,self.size,formats.file_system_format_packetsize(self.formatid))

        if self.use_flanking_primers:
            self.dec = dec_func(self.pf,self.flanking_primer5+self.primer5,\
                                self.flanking_primer3+self.primer3)
        else:
            self.dec = dec_func(self.pf,kwargs['primer5'],kwargs['primer3'])

        for s in self.strands:
            if s.startswith(self.primer5):
                self.dec.decode(s)

        self.dec.write()
        assert self.dec.complete

        self.mem_buffer.seek(0,0) # set read point at beginning of buffer
        
        if self.need_to_close:
            self.in_fd.close()
        return

    def read(self, n=1):        
        return self.mem_buffer.read(n)
    def readline(self, n=-1):
        return self.mem_buffer.readline(n)

    def readable(self):
        return True
    def writable(self):
        return False
        
class WriteDNAFile(DNAFile):
    # WriteDNAFile writes a set of strands.
    def __init__(self,**kwargs):     
        DNAFile.__init__(self)
        if 'formatid' in kwargs:            
            enc_func = formats.file_system_encoder(kwargs['formatid'])
            self.formatid = kwargs['formatid']
        elif 'format_name' in kwargs:
            enc_func = formats.file_system_encoder_by_abbrev(kwargs['format_name'])
            self.formatid = formats.file_system_formatid_by_abbrev(kwargs['format_name'])

        if not ('fsmd_abbrev' in kwargs):
            self.fsmd_abbrev = 'FSMD'
        else:
            self.fsmd_abbrev = kwargs['fsmd_abbrev']

        self.mem_buffer = BytesIO()
        self.pf = ReadPacketizedFilestream(self.mem_buffer)

        assert 'primer5' in kwargs and 'primer3' in kwargs
        self.primer5 = kwargs['primer5']
        self.primer3 = kwargs['primer3']

        if 'flanking_primer3' in kwargs:
            self.flanking_primer3 = kwargs['flanking_primer3']
        else:
            self.flanking_primer3 = ''

        if 'flanking_primer5' in kwargs:
            self.flanking_primer5 = kwargs['flanking_primer5']
        else:
            self.flanking_primer5 = ''

        self.enc = enc_func(self.pf,self.flanking_primer5+self.primer5,\
                            self.flanking_primer3+self.primer3)

            
        if 'output' in kwargs:
            self.output_filename = kwargs['output']
            self.out_fd = open(self.output_filename,"wt")
        elif 'out_fd' in kwargs:
            self.out_fd = kwargs['out_fd']
            self.output_filename = ""
            
        self.size = 0
        self.strands = []
        return

    def _encode_buffer(self):
        for block in self.enc:
            if type(block) == list:
                for s in block:
                    self.strands.append(s)
            else:
                self.strands.append(block)
        #print "_encode_buffer: index={}".format(self.enc.index)

    def read(self, size):
        assert False and "Do not read at the same time as writing"
        
    def writable(self):
        return True
    def readable(self):
        return False
    
    def write(self, buff):
        self.size += len(buff)
        tell = self.mem_buffer.tell()
        self.mem_buffer.seek(0,2)
        try:# fix me
            # convert string
            self.mem_buffer.write(buff)
            self.mem_buffer.seek(tell,0)
        except Exception as e:
            buff = bytearray([x for x in buff])
            self.mem_buffer.write(buff)
            self.mem_buffer.seek(tell,0)
            
        return
    
    def flush(self):
        self._encode_buffer()
        return

    def close(self):
        self.flush()
        hdr = header.encode_file_header(self.output_filename,self.formatid,self.size,\
                                 "",self.flanking_primer5+self.primer5,\
                                 self.flanking_primer3+self.primer3,\
                                 fsmd_abbrev=self.fsmd_abbrev)

        for i,h in enumerate(hdr):
            self.strands.insert(i,h)

        comment = header.encode_file_header_comments(self.output_filename,self.formatid,self.size,"",\
                                              self.primer5,self.primer3)
        self.out_fd.write(comment)
        for s in self.strands:
            self.out_fd.write("{}\n".format(s))

        if self.out_fd != sys.stdout and self.out_fd != sys.stderr:
            self.out_fd.close()
        return


class SegmentedWriteDNAFile(WriteDNAFile):
    # SegmentedWriteDNAFile writes a set of strands.
    def __init__(self,**kwargs):     
        WriteDNAFile.__init__(self,**kwargs)
        self.segments = []
        self.beginIndex = 0
        return

    def _record_segment(self):
        self.segments += [[ self.formatid, self.size, self.primer5, self.primer3, self.beginIndex, self.flanking_primer5, self.flanking_primer3 ]]

    def new_segment(self, format_name, primer5, primer3, flanking_primer5="", flanking_primer3=""):
        self._encode_buffer()  # write everything in the buffer to the file
        self._record_segment() # remember new segment
        self.primer5 = primer5
        self.primer3 = primer3
        self.flanking_primer5 = flanking_primer5
        self.flanking_primer3 = flanking_primer3
        # get current index + 1
        self.beginIndex = self.enc.index
        #print "beginIndex={}".format(self.enc.index)
        enc_func = formats.file_system_encoder_by_abbrev(format_name)
        self.formatid = formats.file_system_formatid_by_abbrev(format_name)
        # we consumed the prior buffer, so just make a new one to avoid
        # cursor positioning problems (JMT: not sure if this is the best way)
        self.mem_buffer = BytesIO()
        self.pf = ReadPacketizedFilestream(self.mem_buffer)
        self.enc = enc_func(self.pf,flanking_primer5+primer5,flanking_primer3+primer3,self.beginIndex)
        self.size=0

    def encode_segments_header(self,segments):
        oprimer5 = segments[0][2]
        oprimer3 = segments[0][3]
        assert len(segments) <= 256
        #if len(segments) == 0:
        #    return []
        hdr = [ len(segments) ]
        logger.debug("encode segments header : {}".format(hdr))
        for s in segments:
            hdr += convertIntToBytes(s[0],2)
            hdr += header.encode_size_and_value( s[1] )
            hdr += header.encode_size_and_value( s[4] )
            hdr += header.encode_primer_diff(oprimer5,s[2])
            hdr += header.encode_primer_diff(oprimer3,s[3])

        return hdr

    def encode_segments_header_comments(self,segments):
        comment = "% segment descriptions\n"        
        tab = "%    "
        for i,s in enumerate(segments):
            comment += tab + "{}. ".format(i) + formats.file_system_format_description(s[0]) + "\n"
            comment += tab + "    size = {}".format(s[1]) + "\n"
            comment += tab + "    5' = {}".format(s[2]) + "\n"
            comment += tab + "    3' = {}".format(s[3]) + "\n"
            comment += tab + "    beginIndex = {}".format(s[4]) + "\n"            
            comment += "%\n"
        return comment

    def close(self):
        logger.debug("WriteSegmentedDNAFile.close")
        
        self.flush()
        self._record_segment() # record last segment
        
        hdr_other = self.encode_segments_header(self.segments)

        formatid = formats.file_system_formatid_by_abbrev("Segmented")

        #print "formatid=",formatid
        
        size = sum([x[1] for x in self.segments])
        primer5 = self.segments[0][2]
        primer3 = self.segments[0][3]
        flanking5 = self.segments[0][-2]
        flanking3 = self.segments[0][-1]
        
        hdr = header.encode_file_header(self.output_filename,formatid,size,hdr_other,flanking5+primer5,flanking3+primer3,fsmd_abbrev=self.fsmd_abbrev)

        #print "Number of strands in header: ", len(hdr)

        for i,h in enumerate(hdr):
            self.strands.insert(i,h)

        comment = header.encode_file_header_comments(self.output_filename,formatid,\
                                              size,hdr_other,primer5,primer3)
        self.out_fd.write(comment)
        comment = self.encode_segments_header_comments(self.segments)
        self.out_fd.write(comment)
        for ss in self.strands:
            if type(ss) is list:
                for s in ss:
                    self.out_fd.write("{}\n".format(s))
            else:
                self.out_fd.write("{}\n".format(ss))
                    
        if self.out_fd != sys.stdout and self.out_fd != sys.stderr:
            self.out_fd.close()
        return


class SegmentedReadDNAFile(ReadDNAFile):

    def decode_segments_header(self,other_data):
        val = other_data
        numSeg = val[0]
        if numSeg==0:
            return
        pos = 1
        allSegs = []
        for i in range(numSeg):
            # get format of this segment
            seg = [convertBytesToInt(val[pos:pos+2])]
            pos+=2

            # get size in bytes
            v,p = header.decode_size_and_value(val,pos)
            pos += p
            seg += [v]

            # get begin index
            v,p = header.decode_size_and_value(val,pos)
            pos += p
            seg += [v]

            primer5,p = header.decode_primer_diff(val[pos:], self.primer5)

            pos += p
            primer3,p = header.decode_primer_diff(val[pos:], self.primer3)

            if i > 0 and self.reverse_primer3_from_seq:
                primer3 = reverse_complement(primer3)
                
            pos += p
            seg.append(primer5)
            seg.append(primer3)
            allSegs.append(seg)

        return allSegs



    # ReadDNAFile reads a set of strands from a file.  It finds the header,
    # determines compatibility and encoding type, and then decodes the file.
    #    
    def __init__(self,**kwargs):     
        ReadDNAFile.__init__(self,**kwargs)

        if 'reverse_primer3_from_seq' in kwargs:
            self.reverse_primer3_from_seq = kwargs['reverse_primer3_from_seq']
        else:
            self.reverse_primer3_from_seq = False
        
        logger.debug("sizeof other_data = {}".format(len(self.header['other_data'])))
        
        if len(self.header['other_data'])==0:
            return

        # restore cursor to end of buffer for writing
        self.mem_buffer.seek(0,2)
        
        segs = self.decode_segments_header(self.header['other_data'])
        self.segments = segs

        if 'preview_mode' in kwargs:
            self.preview_mode = kwargs['preview_mode']
        else:
            self.preview_mode = False
            
        #print "segments=",segs

        for s in segs:
            logger.debug("formatid={} size={} bindex={} primer5={} primer3={}".format(s[0],s[1],s[2],s[3],s[4]))                                             
            formatid = s[0]
            size = s[1]
            bindex = s[2]
            primer5 = s[3]
            primer3 = s[4]

            logger.info("{}".format(s))
            
            if 'use_single_primer' in kwargs and kwargs['use_single_primer']==True:
                # strands from sequencing should all have the same primer
                primer5 = self.primer5
                primer3 = self.primer3

            dec_func = formats.file_system_decoder(formatid)
            #self.mem_buffer = BytesIO()
            self.pf = WritePacketizedFilestream(self.mem_buffer,size,\
                                                formats.file_system_format_packetsize(formatid),\
                                                minKey=bindex)

            #print primer5, primer3, bindex
            self.dec = dec_func(self.pf,primer5,primer3,bindex)

            for s in self.strands:
                #if s.find(primer5)!=-1:
                #print "dnafile.py",self.dec.decode_from_phys_to_strand(s)
                self.dec.decode(s)

            self.dec._attempt_final_decoding()
            #print(self.dec.strand_errors,self.dec.block_errors," missing=",self.dec._packetizedFile.hasMissingKeys())
                            
            if self.preview_mode and (self.dec.block_errors > 0 or not self.dec.complete):
                print("found errors!",self.dec.block_errors)
                print("mising=",self.dec._packetizedFile.hasMissingKeys())
                break
            else:
                self.dec.only_write()
                write_anyway = 'write_incomplete_file' in kwargs and \
                    kwargs['write_incomplete_file']==True
                #if self.dec.complete or write_anyway:
                #    self.dec.write()
                if not write_anyway:
                    assert self.dec.complete

        #print [x for x in self.mem_buffer.getvalue()]
        self.mem_buffer.seek(0,0) # set read point at beginning of buffer

        if self.need_to_close:
            self.in_fd.close()
            
        return

    def read(self, n=1):        
        return self.mem_buffer.read(n)
    def readline(self, n=-1):
        return self.mem_buffer.readline(n)

    def readable(self):
        return True
    def writable(self):
        return False    



from dnastorage.codec.base_codec import BaseCodec
from dnastorage.codec.base_conversion import convertBase, convertFromBase

def binary_encode(packet):
    array = bytearray(packet)
    strand = []
    for b in array:
        strand.append( convertBase(2,b,8) )
    return "".join(strand)


def binary_decode(strand):
    i = 0
    array = []
    while i < len(strand):
        s = strand[i:i+8]
        byte = convertFromBase(2,s)
        array.append(chr(byte))
        i+=len(s)
    packet = bytearray(array)
    return packet


def binary_rotate_encode(strand,prev='A'):
    values = { 'A' : 0 ,
               'C' : 1 ,
               'G' : 2 ,
               'T' : 3   }
    complement = { 'A' : 'T' ,
               'C' : 'G' ,
               'G' : 'C' ,
               'T' : 'A'   }
    counts = { 'A':0, 'G':0,'C':0, 'T':0}
    l = [s for s in strand]
    for i in range(0,len(l)):
        swap = False
        if counts[l[i]] > counts[complement[l[i]]]:
            swap = True
        elif prev == l[i]:
            swap = True
        if swap:
            counts[complement[l[i]]]+=1
            l[i] = complement[l[i]]
            prev = l[i]
        else:
            counts[l[i]] += 1
            prev = l[i]
    #print("".join(l))
    return "".join(l)

def binary_unrotate_decode(strand):
    values = { 'A' : 0 ,
               'C' : 1 ,
               'G' : 2 ,
               'T' : 3   }
    complement = { 'A' : 'A' ,
               'C' : 'C' ,
               'G' : 'C' ,
               'T' : 'A'   }

    l = [s for s in strand]
    for i in range(0,len(l)):
        l[i] = complement[l[i]]
    return "".join(l)

class BinaryCodec(BaseCodec):
    def __init__(self,CodecObj=None,keyWidth=20):
        BaseCodec.__init__(self,CodecObj)
        self._keyWidth=keyWidth

    def _encode(self,packet):
        key = convertBase(2,packet[0],self._keyWidth)
        value = binary_encode(packet[1])
        return key+value

    def _decode(self,s):
        key = convertFromBase(2,s[0:self._keyWidth])
        value = binary_decode(s[self._keyWidth:])
        return key,value

class BinaryRotateCodec(BaseCodec):
    def __init__(self,CodecObj=None,prev='A'):
        BaseCodec.__init__(self,CodecObj)
        self._prev = prev

    def _encode(self,s):
        assert isinstance(s,str) or "Didn't get expected type."
        return binary_rotate_encode(s,self._prev)

    def _decode(self,s):
        assert isinstance(s,str) or "Didn't get expected type."
        return binary_unrotate_decode(s)







from dnastorage.exceptions import NoTolerance, AllowAll
from dnastorage.codec import commafreecodec
from dnastorage.codec.builder import customize_RS_CFC8

def ENC_FSMD_200(pf, primer5, primer3, bIndex=0, policy=NoTolerance(),withCut=None):
    enc = customize_RS_CFC8(True,pf,primer5,primer3,1,0,2,15,90,policy,withCut=withCut,outerECCStrands=20,minIndex=bIndex)
    return enc    

def DEC_FSMD_200(pf, primer5, primer3, bIndex=0,policy=AllowAll(),withCut=None):
    dec = customize_RS_CFC8(False,pf,primer5,primer3,1,0,2,15,90,policy,withCut=withCut,outerECCStrands=20,minIndex=bIndex)
    return dec

def ENC_RS_CFC8_200(pf, primer5, primer3, bIndex=0, policy=NoTolerance()):
    enc = customize_RS_CFC8(True,pf,primer5,primer3,1,2,2,15,185*15,policy,minIndex=bIndex)
    return enc    

def DEC_RS_CFC8_200(pf, primer5, primer3, bIndex=0, policy=AllowAll()):
    dec = customize_RS_CFC8(False,pf,primer5,primer3,1,2,2,15,185*15,policy,minIndex=bIndex)
    return dec    

### Support for preview - FIXME - move to dnapreview - create registry

def ENC_FSMD_WCUT_160(pf, primer5, primer3, bIndex=0, policy=NoTolerance(),withCut=None):
    withCut="AGGTACCA"
    enc = customize_RS_CFC8(True,pf,primer5,primer3,1,0,2,10,120,policy,withCut=withCut,outerECCStrands=20,minIndex=bIndex)
    return enc    

def DEC_FSMD_WCUT_160(pf, primer5, primer3, bIndex=0,policy=AllowAll(),withCut=None):
    withCut="AGGTACCA"
    dec = customize_RS_CFC8(False,pf,primer5,primer3,1,0,2,10,120,policy,withCut=withCut,outerECCStrands=20,minIndex=bIndex)
    return dec

def ENC_RS_CFC8_RE1_160(pf, primer5, primer3, bIndex=0, policy=NoTolerance()):
    withCut="AGGTACCA"
    enc = customize_RS_CFC8(True,pf,primer5,primer3,1,1,3,9,185*9,policy,minIndex=bIndex,
                            withCut=withCut)
    return enc    

def DEC_RS_CFC8_RE1_160(pf, primer5, primer3, bIndex=0, policy=AllowAll()):
    withCut="AGGTACCA"
    dec = customize_RS_CFC8(False,pf,primer5,primer3,1,1,3,9,185*9,policy,minIndex=bIndex,\
                            withCut=withCut)
    return dec    

def ENC_RS_CFC8_RE2_160(pf, primer5, primer3, bIndex=0, policy=NoTolerance()):
    withCut="CCTGCAGG"
    enc = customize_RS_CFC8(True,pf,primer5,primer3,1,1,3,9,185*9,policy,minIndex=bIndex,
                            withCut=withCut)
    return enc    

def DEC_RS_CFC8_RE2_160(pf, primer5, primer3, bIndex=0, policy=AllowAll()):
    withCut="CCTGCAGG"
    dec = customize_RS_CFC8(False,pf,primer5,primer3,1,1,3,9,185*9,policy,minIndex=bIndex,\
                            withCut=withCut)
    return dec    

def ENC_RS_CFC8_RE3_160(pf, primer5, primer3, bIndex=0, policy=NoTolerance()):
    withCut="GCGGCCGC"
    enc = customize_RS_CFC8(True,pf,primer5,primer3,1,1,3,9,185*9,policy,minIndex=bIndex,
                            withCut=withCut)
    return enc    

def DEC_RS_CFC8_RE3_160(pf, primer5, primer3, bIndex=0, policy=AllowAll()):
    withCut="GCGGCCGC"
    dec = customize_RS_CFC8(False,pf,primer5,primer3,1,1,3,9,185*9,policy,minIndex=bIndex,\
                            withCut=withCut)
    return dec    

def ENC_RS_CFC8_RE4_160(pf, primer5, primer3, bIndex=0, policy=NoTolerance()):
    withCut="GTTTAAAC"
    enc = customize_RS_CFC8(True,pf,primer5,primer3,1,1,3,9,185*9,policy,minIndex=bIndex,
                            withCut=withCut)
    return enc    

def DEC_RS_CFC8_RE4_160(pf, primer5, primer3, bIndex=0, policy=AllowAll()):
    withCut="GTTTAAAC"
    dec = customize_RS_CFC8(False,pf,primer5,primer3,1,1,3,9,185*9,policy,minIndex=bIndex,\
                            withCut=withCut)
    return dec    

### End support for preview - FIXME - move to dnapreview - create registry

FileSystemFormats = {
    # KEY      KEY    LEN  PacketSize, Abbrev.   Description                   ENCODER       DECODER
    0x0010 : [0x0010, 200, 90, "FSMD", "File system meta-data format", ENC_FSMD_200, DEC_FSMD_200 ],
    0x0020 : [0x0020, 200, 16, "RS+CFC8", "Reed-Solomon coded with ad hoc Comma-free codewords",
              ENC_RS_CFC8_200, DEC_RS_CFC8_200 ],

    # Support preview -- support through registry
    0x0011 : [0x0011, 160, 120, "FSMD-1", "File system meta-data format with cut", ENC_FSMD_WCUT_160, DEC_FSMD_WCUT_160 ],
    0x1000 : [0x1000, 200, 20, "Segmented", "Segmented file format to support Preview", None, None],
    0x2021 : [0x2021, 160, 9, "RS+CFC8+RE1", "Reed-Solomon coded with Comma-free codewords",
              ENC_RS_CFC8_RE1_160, DEC_RS_CFC8_RE1_160 ],
    0x2022 : [0x2022, 160, 9, "RS+CFC8+RE2", "Reed-Solomon coded with Comma-free codewords",
              ENC_RS_CFC8_RE2_160, DEC_RS_CFC8_RE2_160 ],
    0x2023 : [0x2023, 160, 9, "RS+CFC8+RE3", "Reed-Solomon coded with Comma-free codewords",
              ENC_RS_CFC8_RE3_160, DEC_RS_CFC8_RE3_160 ],
    0x2024 : [0x2024, 160, 9, "RS+CFC8+RE4", "Reed-Solomon coded with Comma-free codewords",
              ENC_RS_CFC8_RE4_160, DEC_RS_CFC8_RE4_160 ],
}

def file_system_formats():
    return [ v[3] for k,v in FileSystemFormats.items() ]

_abbrevFileSystemDict = { v[3] : v for k,v in FileSystemFormats.items() }

def file_system_format_description(formatid):
    return FileSystemFormats[formatid][4]

def file_system_format_packetsize(formatid):
    return FileSystemFormats[formatid][2]

def file_system_encoder(formatid):
    return FileSystemFormats[formatid][5]

def file_system_decoder(formatid):
    return FileSystemFormats[formatid][6]

def file_system_encoder_by_abbrev(ab):
    return _abbrevFileSystemDict[ab][5]

def file_system_decoder_by_abbrev(ab):
    return _abbrevFileSystemDict[ab][6]

def file_system_formatid_by_abbrev(ab):
    return _abbrevFileSystemDict[ab][0]






import random
from dnastorage.codec.base_codec import BaseCodec

dense_enc_table = [
'AAAA','AAAC','AAAT','AAAG','AACA','AACC','AACT','AACG',
'AATA','AATC','AATT','AATG','AAGA','AAGC','AAGT','AAGG',
'ACAA','ACAC','ACAT','ACAG','ACCA','ACCC','ACCT','ACCG',
'ACTA','ACTC','ACTT','ACTG','ACGA','ACGC','ACGT','ACGG',
'ATAA','ATAC','ATAT','ATAG','ATCA','ATCC','ATCT','ATCG',
'ATTA','ATTC','ATTT','ATTG','ATGA','ATGC','ATGT','ATGG',
'AGAA','AGAC','AGAT','AGAG','AGCA','AGCC','AGCT','AGCG',
'AGTA','AGTC','AGTT','AGTG','AGGA','AGGC','AGGT','AGGG',
'CAAA','CAAC','CAAT','CAAG','CACA','CACC','CACT','CACG',
'CATA','CATC','CATT','CATG','CAGA','CAGC','CAGT','CAGG',
'CCAA','CCAC','CCAT','CCAG','CCCA','CCCC','CCCT','CCCG',
'CCTA','CCTC','CCTT','CCTG','CCGA','CCGC','CCGT','CCGG',
'CTAA','CTAC','CTAT','CTAG','CTCA','CTCC','CTCT','CTCG',
'CTTA','CTTC','CTTT','CTTG','CTGA','CTGC','CTGT','CTGG',
'CGAA','CGAC','CGAT','CGAG','CGCA','CGCC','CGCT','CGCG',
'CGTA','CGTC','CGTT','CGTG','CGGA','CGGC','CGGT','CGGG',
'TAAA','TAAC','TAAT','TAAG','TACA','TACC','TACT','TACG',
'TATA','TATC','TATT','TATG','TAGA','TAGC','TAGT','TAGG',
'TCAA','TCAC','TCAT','TCAG','TCCA','TCCC','TCCT','TCCG',
'TCTA','TCTC','TCTT','TCTG','TCGA','TCGC','TCGT','TCGG',
'TTAA','TTAC','TTAT','TTAG','TTCA','TTCC','TTCT','TTCG',
'TTTA','TTTC','TTTT','TTTG','TTGA','TTGC','TTGT','TTGG',
'TGAA','TGAC','TGAT','TGAG','TGCA','TGCC','TGCT','TGCG',
'TGTA','TGTC','TGTT','TGTG','TGGA','TGGC','TGGT','TGGG',
'GAAA','GAAC','GAAT','GAAG','GACA','GACC','GACT','GACG',
'GATA','GATC','GATT','GATG','GAGA','GAGC','GAGT','GAGG',
'GCAA','GCAC','GCAT','GCAG','GCCA','GCCC','GCCT','GCCG',
'GCTA','GCTC','GCTT','GCTG','GCGA','GCGC','GCGT','GCGG',
'GTAA','GTAC','GTAT','GTAG','GTCA','GTCC','GTCT','GTCG',
'GTTA','GTTC','GTTT','GTTG','GTGA','GTGC','GTGT','GTGG',
'GGAA','GGAC','GGAT','GGAG','GGCA','GGCC','GGCT','GGCG',
'GGTA','GGTC','GGTT','GGTG','GGGA','GGGC','GGGT','GGGG'
]

dense_dec_table = {
'AAAA' : 0,
'AAAC' : 1,
'AAAT' : 2,
'AAAG' : 3,
'AACA' : 4,
'AACC' : 5,
'AACT' : 6,
'AACG' : 7,
'AATA' : 8,
'AATC' : 9,
'AATT' : 10,
'AATG' : 11,
'AAGA' : 12,
'AAGC' : 13,
'AAGT' : 14,
'AAGG' : 15,
'ACAA' : 16,
'ACAC' : 17,
'ACAT' : 18,
'ACAG' : 19,
'ACCA' : 20,
'ACCC' : 21,
'ACCT' : 22,
'ACCG' : 23,
'ACTA' : 24,
'ACTC' : 25,
'ACTT' : 26,
'ACTG' : 27,
'ACGA' : 28,
'ACGC' : 29,
'ACGT' : 30,
'ACGG' : 31,
'ATAA' : 32,
'ATAC' : 33,
'ATAT' : 34,
'ATAG' : 35,
'ATCA' : 36,
'ATCC' : 37,
'ATCT' : 38,
'ATCG' : 39,
'ATTA' : 40,
'ATTC' : 41,
'ATTT' : 42,
'ATTG' : 43,
'ATGA' : 44,
'ATGC' : 45,
'ATGT' : 46,
'ATGG' : 47,
'AGAA' : 48,
'AGAC' : 49,
'AGAT' : 50,
'AGAG' : 51,
'AGCA' : 52,
'AGCC' : 53,
'AGCT' : 54,
'AGCG' : 55,
'AGTA' : 56,
'AGTC' : 57,
'AGTT' : 58,
'AGTG' : 59,
'AGGA' : 60,
'AGGC' : 61,
'AGGT' : 62,
'AGGG' : 63,
'CAAA' : 64,
'CAAC' : 65,
'CAAT' : 66,
'CAAG' : 67,
'CACA' : 68,
'CACC' : 69,
'CACT' : 70,
'CACG' : 71,
'CATA' : 72,
'CATC' : 73,
'CATT' : 74,
'CATG' : 75,
'CAGA' : 76,
'CAGC' : 77,
'CAGT' : 78,
'CAGG' : 79,
'CCAA' : 80,
'CCAC' : 81,
'CCAT' : 82,
'CCAG' : 83,
'CCCA' : 84,
'CCCC' : 85,
'CCCT' : 86,
'CCCG' : 87,
'CCTA' : 88,
'CCTC' : 89,
'CCTT' : 90,
'CCTG' : 91,
'CCGA' : 92,
'CCGC' : 93,
'CCGT' : 94,
'CCGG' : 95,
'CTAA' : 96,
'CTAC' : 97,
'CTAT' : 98,
'CTAG' : 99,
'CTCA' : 100,
'CTCC' : 101,
'CTCT' : 102,
'CTCG' : 103,
'CTTA' : 104,
'CTTC' : 105,
'CTTT' : 106,
'CTTG' : 107,
'CTGA' : 108,
'CTGC' : 109,
'CTGT' : 110,
'CTGG' : 111,
'CGAA' : 112,
'CGAC' : 113,
'CGAT' : 114,
'CGAG' : 115,
'CGCA' : 116,
'CGCC' : 117,
'CGCT' : 118,
'CGCG' : 119,
'CGTA' : 120,
'CGTC' : 121,
'CGTT' : 122,
'CGTG' : 123,
'CGGA' : 124,
'CGGC' : 125,
'CGGT' : 126,
'CGGG' : 127,
'TAAA' : 128,
'TAAC' : 129,
'TAAT' : 130,
'TAAG' : 131,
'TACA' : 132,
'TACC' : 133,
'TACT' : 134,
'TACG' : 135,
'TATA' : 136,
'TATC' : 137,
'TATT' : 138,
'TATG' : 139,
'TAGA' : 140,
'TAGC' : 141,
'TAGT' : 142,
'TAGG' : 143,
'TCAA' : 144,
'TCAC' : 145,
'TCAT' : 146,
'TCAG' : 147,
'TCCA' : 148,
'TCCC' : 149,
'TCCT' : 150,
'TCCG' : 151,
'TCTA' : 152,
'TCTC' : 153,
'TCTT' : 154,
'TCTG' : 155,
'TCGA' : 156,
'TCGC' : 157,
'TCGT' : 158,
'TCGG' : 159,
'TTAA' : 160,
'TTAC' : 161,
'TTAT' : 162,
'TTAG' : 163,
'TTCA' : 164,
'TTCC' : 165,
'TTCT' : 166,
'TTCG' : 167,
'TTTA' : 168,
'TTTC' : 169,
'TTTT' : 170,
'TTTG' : 171,
'TTGA' : 172,
'TTGC' : 173,
'TTGT' : 174,
'TTGG' : 175,
'TGAA' : 176,
'TGAC' : 177,
'TGAT' : 178,
'TGAG' : 179,
'TGCA' : 180,
'TGCC' : 181,
'TGCT' : 182,
'TGCG' : 183,
'TGTA' : 184,
'TGTC' : 185,
'TGTT' : 186,
'TGTG' : 187,
'TGGA' : 188,
'TGGC' : 189,
'TGGT' : 190,
'TGGG' : 191,
'GAAA' : 192,
'GAAC' : 193,
'GAAT' : 194,
'GAAG' : 195,
'GACA' : 196,
'GACC' : 197,
'GACT' : 198,
'GACG' : 199,
'GATA' : 200,
'GATC' : 201,
'GATT' : 202,
'GATG' : 203,
'GAGA' : 204,
'GAGC' : 205,
'GAGT' : 206,
'GAGG' : 207,
'GCAA' : 208,
'GCAC' : 209,
'GCAT' : 210,
'GCAG' : 211,
'GCCA' : 212,
'GCCC' : 213,
'GCCT' : 214,
'GCCG' : 215,
'GCTA' : 216,
'GCTC' : 217,
'GCTT' : 218,
'GCTG' : 219,
'GCGA' : 220,
'GCGC' : 221,
'GCGT' : 222,
'GCGG' : 223,
'GTAA' : 224,
'GTAC' : 225,
'GTAT' : 226,
'GTAG' : 227,
'GTCA' : 228,
'GTCC' : 229,
'GTCT' : 230,
'GTCG' : 231,
'GTTA' : 232,
'GTTC' : 233,
'GTTT' : 234,
'GTTG' : 235,
'GTGA' : 236,
'GTGC' : 237,
'GTGT' : 238,
'GTGG' : 239,
'GGAA' : 240,
'GGAC' : 241,
'GGAT' : 242,
'GGAG' : 243,
'GGCA' : 244,
'GGCC' : 245,
'GGCT' : 246,
'GGCG' : 247,
'GGTA' : 248,
'GGTC' : 249,
'GGTT' : 250,
'GGTG' : 251,
'GGGA' : 252,
'GGGC' : 253,
'GGGT' : 254,
'GGGG' : 255
}

def dense_encode_byte(byte):
    return dense_enc_table[byte]

def dense_decode_byte(byte):
    assert byte in dense_dec_table
    return dense_dec_table[byte]

def dense_encode(packet):
    array = bytearray(packet)
    strand = []
    for b in array:
        strand.append( dense_encode_byte(b) )
    return "".join(strand)

def dense_decode(strand):
    i = 0
    array = []
    while i < len(strand):
        s = strand[i:i+4]
        byte = dense_decode_byte(s)
        array.append(byte)
        i+=4
    packet = bytearray(array)
    return packet

class DenseCodewords(BaseCodec):
    def __init__(self,numberSymbols,CodecObj=None):
        BaseCodec.__init__(self,CodecObj)
        self._numberBytes=numberSymbols

    def _encode(self,strand):
        enc_strand = dense_encode(strand)
        return enc_strand

    def _decode(self,s):
        value = dense_decode(s)
        return value

